"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import logging
import os
from abc import ABCMeta, abstractmethod
from glob import glob
from typing import TYPE_CHECKING, TypeVar

import numpy as np
import pandas as pd

from fairchem.core.common import distutils
from fairchem.core.common.logger import WandBSingletonLogger
from fairchem.core.components.calculate._calculate_runner import (
    CalculateRunner,
)
from fairchem.core.components.reducer import Reducer

if TYPE_CHECKING:
    from collections.abc import Sequence

R = TypeVar("R")
M = TypeVar("M")


class BenchmarkReducer(Reducer, metaclass=ABCMeta):
    """Benchmark reducer interface class.

    Note:
        When running with the `fairchemv2` cli, the `job_config` and `runner_config` attributes are set at
        runtime to those given in the config file.

    See the Reducer interface class for implementation details.

    Attributes:
        job_config (DictConfig): a managed attribute that gives access to the job config
        runner_config (DictConfig): a managed attributed that gives access to the calling runner config
    """

    @property
    def runner_type(self) -> type[CalculateRunner]:
        """The runner type this reducer is associated with."""
        return CalculateRunner

    @property
    def glob_pattern(self):
        """Returns the glob pattern used to find result files from the runner."""
        return self.runner_type.result_glob_pattern

    @property
    def logger(self) -> WandBSingletonLogger | None:
        """Returns a logger instance if conditions are met, otherwise None.

        Returns:
            WandBSingletonLogger or None: Logger instance if running on main rank with logging enabled
        """
        if (
            distutils.is_master()
            and self.job_config is not None
            and not self.job_config.debug
            and self.job_config.logger
        ):
            return WandBSingletonLogger.get_instance()
        else:
            return None

    @abstractmethod
    def join_results(self, results_dir: str, glob_pattern: str) -> R:
        """Join results from multiple files into a single result object.

        Args:
            results_dir: Directory containing result files
            glob_pattern: Pattern to match result files

        Returns:
            Combined results object of type R
        """
        raise NotImplementedError

    @abstractmethod
    def save_results(self, results: R, results_dir: str) -> None:
        """Save joined results to file

        Args:
            results:  results: Combined results from join_results
            results_dir: Directory containing result files
        """
        raise NotImplementedError

    @abstractmethod
    def compute_metrics(self, results: R, run_name: str) -> M:
        """Compute metrics from the joined results.

        Args:
            results: Combined results from join_results
            run_name: Name of the current run

        Returns:
            Metrics object of type M
        """
        raise NotImplementedError

    @abstractmethod
    def save_metrics(self, metrics: M, results_dir: str) -> None:
        """Save computed metrics to disk.

        Args:
            metrics: Metrics object to save
            results_dir: Directory to save metrics to
        """
        raise NotImplementedError

    @abstractmethod
    def log_metrics(self, metrics: M, run_name: str):
        """Log metrics to the configured logger.

        Args:
            metrics: Metrics object to log
            run_name: Name of the current run
        """
        raise NotImplementedError

    @abstractmethod
    def save_state(self, checkpoint_location: str, is_preemption: bool = False) -> bool:
        """Save the current state of the reducer to a checkpoint.

        Args:
            checkpoint_location: Location to save the checkpoint
            is_preemption: Whether the save is due to preemption

        Returns:
            bool: Success status of the save operation
        """
        raise NotImplementedError

    @abstractmethod
    def load_state(self, checkpoint_location: str | None) -> None:
        """Load reducer state from a checkpoint.

        Args:
            checkpoint_location: Location to load the checkpoint from, or None
        """
        raise NotImplementedError

    def reduce(self):
        """Join results, compute metrics, save and log resulting metrics.

        Note:
            Re-implementing this method in derived classes is discouraged.
        """
        # re-implementing  this method in derived classes is discouraged
        logging.info(
            f"Joining calculation results in {self.job_config.metadata.results_dir}"
        )
        results = self.join_results(
            self.job_config.metadata.results_dir, self.glob_pattern
        )
        logging.info(
            f"Saving joined results in {self.job_config.metadata.results_dir}."
        )
        self.save_results(results, self.job_config.metadata.results_dir)
        logging.info("Calculating metrics.")
        metrics = self.compute_metrics(results, run_name=self.job_config.run_name)
        if metrics is not None:
            logging.info(
                f"Saving computed metrics in {self.job_config.metadata.results_dir}"
            )
            self.save_metrics(metrics, self.job_config.metadata.results_dir)
            if self.logger is not None:
                self.log_metrics(metrics, run_name=self.job_config.run_name)
        else:
            logging.info(
                "Evaluation labels not available, skipping metrics computation. Please submit results to the appropriate benchmark leaderboard if applicable."
            )


class JsonDFReducer(BenchmarkReducer):
    """A common pandas DataFrame reducer for benchmarks

    Results are assumed to be saved as json files that can be read into pandas dataframes.
    Only mean absolute error is computed for common columns in the predicted results and target data
    """

    def __init__(
        self,
        benchmark_name: str,
        target_data_path: str | None = None,
        target_data_keys: Sequence[str] | None = None,
        index_name: str | None = None,
    ):
        """Initialize the JsonDFReducer with benchmark data and configuration.

        Args:
            benchmark_name: Name of the benchmark, used for file naming
            target_data_path: Path to the target data JSON file
            target_data_keys: List of target property keys to extract from results file
            index_name: Optional name of the column to use as index
        """

        if target_data_path is None and target_data_keys is None:
            raise ValueError(
                "Either target_data_path or target_data_keys should be provided"
            )

        self.index_name = index_name
        self.benchmark_name = benchmark_name
        self.target_data = (
            self.load_targets(target_data_path, index_name)
            if target_data_path is not None
            else None
        )
        self.target_data_keys = target_data_keys

    @staticmethod
    def load_targets(path: str, index_name: str | None) -> pd.DataFrame:
        """Load target data from a JSON file into a pandas DataFrame.

        Args:
            path: Path to the target JSON file
            index_name: Optional name of the column to use as index

        Returns:
            DataFrame containing the target data, sorted by index
        """
        df_targets = pd.read_json(path, dtype=False)
        if index_name is not None:
            df_targets = df_targets.set_index(index_name)
        return df_targets.sort_index()

    def join_results(self, results_dir: str, glob_pattern: str) -> pd.DataFrame:
        """Join results from multiple JSON files into a single DataFrame.

        Args:
            results_dir: Directory containing result files
            glob_pattern: Pattern to match result files

        Returns:
            Combined DataFrame containing all results
        """
        results = pd.concat(
            [
                pd.read_json(f, dtype=False)
                for f in glob(os.path.join(results_dir, glob_pattern))
            ]
        ).reset_index()

        if self.index_name is not None:
            results = results.set_index("sid").sort_index()
            results.index.name = self.index_name

        return results

    def save_results(self, results: pd.DataFrame, results_dir: str) -> None:
        """Save joined results to a compressed json file

        Args:
            results:  results: Combined results from join_results
            results_dir: Directory containing result files
        """
        results.reset_index().to_json(
            os.path.join(results_dir, f"{self.benchmark_name}_results.json.gz")
        )

    def compute_metrics(self, results: pd.DataFrame, run_name: str) -> pd.DataFrame:
        """Compute mean absolute error metrics for common columns between results and targets.

        Args:
            results: DataFrame containing prediction results
            run_name: Name of the current run, used as index in the metrics DataFrame

        Returns:
            DataFrame containing computed metrics with run_name as index
        """
        """This will just compute MAE of everything that is common in the results and target dataframes"""

        metrics = {}
        if self.target_data is not None:
            common_cols = [
                col for col in results.columns if col in self.target_data.columns
            ]
            metrics.update(
                {
                    f"{col},mae": (results[col] - self.target_data[col]).abs().mean()
                    for col in common_cols
                }
            )

        if self.target_data_keys is not None:
            for target_name in self.target_data_keys:
                # TODO: For now we'll hardcode forces, but a more general
                # approach for arbitrary metrics + properties should be
                # incorporated
                if target_name == "forces":
                    forces = np.concatenate(results[target_name].values)
                    forces_norm = np.linalg.norm(forces, axis=1)
                    forces_target = np.concatenate(
                        results[f"{target_name}_target"].values
                    )
                    forces_target_norm = np.linalg.norm(forces_target, axis=1)

                    metrics[f"{target_name},mae"] = np.mean(
                        np.abs(forces - forces_target)
                    )
                    metrics[f"{target_name},cosine_similarity"] = np.sum(
                        forces_target * forces
                    ) / max(
                        np.linalg.norm(forces_target) * np.linalg.norm(forces), 1e-8
                    )
                    metrics[f"{target_name},magnitude_error"] = np.mean(
                        np.abs(forces_norm - forces_target_norm)
                    )
                else:
                    metrics[f"{target_name},mae"] = (
                        (results[target_name] - results[f"{target_name}_target"])
                        .abs()
                        .mean()
                    )

        return pd.DataFrame([metrics], index=[run_name])

    def save_metrics(self, metrics: pd.DataFrame, results_dir: str) -> None:
        """Save computed metrics to a compressed JSON file.

        Args:
            metrics: DataFrame containing the computed metrics
            results_dir: Directory where metrics will be saved
        """
        metrics.to_json(
            os.path.join(results_dir, f"{self.benchmark_name}_metrics.json.gz")
        )

    def log_metrics(self, metrics: pd.DataFrame, run_name: str) -> None:
        """Log metrics to the configured logger if available.

        Args:
            metrics: DataFrame containing the computed metrics
            run_name: Name of the current run
        """
        if self.logger is not None:
            self.logger.log_dataframe(
                name=self.benchmark_name, dataframe=metrics.reset_index()
            )

    def save_state(self, checkpoint_location: str, is_preemption: bool = False) -> bool:
        pass

    def load_state(self, checkpoint_location: str | None) -> None:
        pass



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.

Creates LMDB files with extracted graph features from provided *.extxyz files
for the S2EF task.
"""

from __future__ import annotations

import argparse
import glob
import multiprocessing as mp
import os
import pickle

import ase.io
import lmdb
import numpy as np
import torch
from tqdm import tqdm

from fairchem.core.preprocessing import AtomsToGraphs


def write_images_to_lmdb(mp_arg):
    a2g, db_path, samples, sampled_ids, idx, pid, args = mp_arg
    db = lmdb.open(
        db_path,
        map_size=1099511627776 * 2,
        subdir=False,
        meminit=False,
        map_async=True,
    )

    pbar = tqdm(
        total=5000 * len(samples),
        position=pid,
        desc="Preprocessing data into LMDBs",
    )
    for sample in samples:
        with open(sample) as fp:
            traj_logs = fp.read().splitlines()
        xyz_idx = os.path.splitext(os.path.basename(sample))[0]
        traj_path = os.path.join(args.data_path, f"{xyz_idx}.extxyz")
        traj_frames = ase.io.read(traj_path, ":")

        for i, frame in enumerate(traj_frames):
            frame_log = traj_logs[i].split(",")
            sid = int(frame_log[0].split("random")[1])
            fid = int(frame_log[1].split("frame")[1])
            data_object = a2g.convert(frame)
            # add atom tags
            data_object.tags = torch.LongTensor(frame.get_tags())
            data_object.sid = sid
            data_object.fid = fid
            # subtract off reference energy
            if args.ref_energy and not args.test_data:
                ref_energy = float(frame_log[2])
                data_object.energy -= ref_energy

            txn = db.begin(write=True)
            txn.put(
                f"{idx}".encode("ascii"),
                pickle.dumps(data_object, protocol=-1),
            )
            txn.commit()
            idx += 1
            sampled_ids.append(",".join(frame_log[:2]) + "\n")
            pbar.update(1)

    # Save count of objects in lmdb.
    txn = db.begin(write=True)
    txn.put("length".encode("ascii"), pickle.dumps(idx, protocol=-1))
    txn.commit()

    db.sync()
    db.close()

    return sampled_ids, idx


def main(args: argparse.Namespace) -> None:
    xyz_logs = glob.glob(os.path.join(args.data_path, "*.txt"))
    if not xyz_logs:
        raise RuntimeError("No *.txt files found. Did you uncompress?")
    if args.num_workers > len(xyz_logs):
        args.num_workers = len(xyz_logs)

    # Initialize feature extractor.
    a2g = AtomsToGraphs(
        max_neigh=50,
        radius=6,
        r_energy=not args.test_data,
        r_forces=not args.test_data,
        r_fixed=True,
        r_distances=False,
        r_edges=args.get_edges,
    )

    # Create output directory if it doesn't exist.
    os.makedirs(os.path.join(args.out_path), exist_ok=True)

    # Initialize lmdb paths
    db_paths = [
        os.path.join(args.out_path, "data.%04d.lmdb" % i)
        for i in range(args.num_workers)
    ]

    # Chunk the trajectories into args.num_workers splits
    chunked_txt_files = np.array_split(xyz_logs, args.num_workers)

    # Extract features
    sampled_ids, idx = [[]] * args.num_workers, [0] * args.num_workers

    pool = mp.Pool(args.num_workers)
    mp_args = [
        (
            a2g,
            db_paths[i],
            chunked_txt_files[i],
            sampled_ids[i],
            idx[i],
            i,
            args,
        )
        for i in range(args.num_workers)
    ]
    op = list(zip(*pool.imap(write_images_to_lmdb, mp_args)))
    sampled_ids, idx = list(op[0]), list(op[1])

    # Log sampled image, trajectory trace
    for j, i in enumerate(range(args.num_workers)):
        with open(os.path.join(args.out_path, "data_log.%04d.txt" % i), "w") as ids_log:
            ids_log.writelines(sampled_ids[j])


def get_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--data-path",
        help="Path to dir containing *.extxyz and *.txt files",
    )
    parser.add_argument(
        "--out-path",
        help="Directory to save extracted features. Will create if doesn't exist",
    )
    parser.add_argument(
        "--get-edges",
        action="store_true",
        help="Store edge indices in LMDB, ~10x storage requirement. Default: compute edge indices on-the-fly.",
    )
    parser.add_argument(
        "--num-workers",
        type=int,
        default=1,
        help="No. of feature-extracting processes or no. of dataset chunks",
    )
    parser.add_argument(
        "--ref-energy", action="store_true", help="Subtract reference energies"
    )
    parser.add_argument(
        "--test-data",
        action="store_true",
        help="Is data being processed test data?",
    )
    return parser


if __name__ == "__main__":
    parser: argparse.ArgumentParser = get_parser()
    args: argparse.Namespace = parser.parse_args()
    main(args)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import math

import torch


class SO3_Linear(torch.nn.Module):
    def __init__(self, in_features: int, out_features: int, lmax: int) -> None:
        """
        1. Use `torch.einsum` to prevent slicing and concatenation
        2. Need to specify some behaviors in `no_weight_decay` and weight initialization.
        """
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.lmax = lmax

        self.weight = torch.nn.Parameter(
            torch.randn((self.lmax + 1), out_features, in_features)
        )
        bound = 1 / math.sqrt(self.in_features)
        torch.nn.init.uniform_(self.weight, -bound, bound)
        self.bias = torch.nn.Parameter(torch.zeros(out_features))

        expand_index = torch.zeros([(lmax + 1) ** 2]).long()
        for lval in range(lmax + 1):
            start_idx = lval**2
            length = 2 * lval + 1
            expand_index[start_idx : (start_idx + length)] = lval
        self.register_buffer("expand_index", expand_index, persistent=False)

    def forward(self, input_embedding):
        weight = torch.index_select(
            self.weight, dim=0, index=self.expand_index
        )  # [(L_max + 1) ** 2, C_out, C_in]
        out = torch.einsum(
            "bmi, moi -> bmo", input_embedding, weight
        )  # [N, (L_max + 1) ** 2, C_out]
        bias = self.bias.view(1, 1, self.out_features)
        out[:, 0:1, :] = out.narrow(1, 0, 1) + bias
        return out

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(in_features={self.in_features}, out_features={self.out_features}, lmax={self.lmax})"



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.

OMol Evaluation Recipes
========================

This module provides evaluation recipes for various molecular property tasks proposed in the OMol25 paper.

The module includes functions for:
- Geometry optimizations and conformer generation
- Protonation state energetics
- Ionization energies and electron affinities
- Spin gap calculations
- Protein-ligand interactions
- Ligand strain energies
- Distance scaling behavior
- Single-point energy and force calculations

Each function follows a consistent pattern of taking input data and any ASE calculator,
performing the required calculations, and returning results in a standardized format
suitable for downstream evaluation on the OMol leaderboard - #TODO: add link.
"""

from __future__ import annotations

import logging
from collections import defaultdict
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from ase import Atoms
    from ase.calculators.calculator import Calculator

from fairchem.data.omol.orca.calc import EVAL_OPT_PARAMETERS
from pymatgen.io.ase import MSONAtoms
from tqdm import tqdm


def relax_job(
    atoms: Atoms, calculator: Calculator, opt_flags: dict[str, Any]
) -> dict[str, Any]:
    """
    Perform a geometry optimization job on an atomic structure.

    This function optimizes the geometry of an atomic structure using the provided
    calculator and optimization parameters. It captures both the initial and final
    states (energy, forces, and atomic positions) for comparison.

    Args:
        atoms: ASE Atoms object representing the initial structure
        calculator: ASE calculator object (e.g., FAIRChemCalculator) to use for energy/force calculations
        opt_flags (dict): Dictionary containing optimization parameters including:
            - optimizer: ASE optimizer class (e.g., BFGS, FIRE)
            - optimizer_kwargs: Additional kwargs for the optimizer
            - fmax: Force convergence criterion (eV/Å)
            - max_steps: Maximum number of optimization steps

    Returns:
        dict: Results organized in the following form -
        {
            "initial": {
                "atoms": MSONAtoms dictionary of initial structure,
                "energy": Initial total energy (eV),
                "forces": Initial forces as a list (eV/Å),
            },
            "final": {
                "atoms": MSONAtoms dictionary of optimized structure,
                "energy": Final total energy (eV),
                "forces": Final forces as a list (eV/Å),
            }
        }

    Note:
        If optimization fails, the function logs the error and returns the last
        valid state rather than crashing.
    """
    atoms.calc = calculator
    initial_energy = atoms.get_potential_energy()
    initial_forces = atoms.get_forces()
    initial_atoms = atoms.copy()

    try:
        dyn = opt_flags["optimizer"](atoms, **opt_flags["optimizer_kwargs"])
        dyn.run(fmax=opt_flags["fmax"], steps=opt_flags["max_steps"])
    except Exception as e:
        # atoms are updated in place, so no actual change needed.
        logging.info(f"Optimization failed, using last valid step. {e}")
    energy = atoms.get_potential_energy()
    forces = atoms.get_forces()
    result = {
        "initial": {
            "atoms": MSONAtoms(initial_atoms).as_dict(),
            "energy": initial_energy,
            "forces": initial_forces.tolist(),
        },
        "final": {
            "atoms": MSONAtoms(atoms).as_dict(),
            "energy": energy,
            "forces": forces.tolist(),
        },
    }
    return result


def single_point_job(atoms: Atoms, calculator: Calculator) -> dict[str, Any]:
    """
    Perform a single-point energy and force calculation.

    This function calculates the energy and forces for a given atomic structure.

    Args:
        atoms: ASE Atoms object representing the structure
        calculator: ASE calculator object (e.g., FAIRChemCalculator) to use for energy/force calculations

    Returns:
        dict: Results organized in the following form -
        {
            "atoms": MSONAtoms dictionary of the structure,
            "energy": Total energy (eV),
            "forces": Forces as a list (eV/Å),
        }
    """
    # clear cache, especially for ieea + spin-gap
    calculator.reset()
    atoms.calc = calculator
    energy = atoms.get_potential_energy()
    forces = atoms.get_forces()
    atoms.calc = None
    result = {
        "atoms": MSONAtoms(atoms).as_dict(),
        "energy": energy,
        "forces": forces.tolist(),
    }
    return result


def conformers(input_data: dict[str, Any], calculator: Calculator) -> dict[str, Any]:
    """
    Calculate conformer energies and geometries.

    This function performs geometry optimizations on molecular conformers.

    Args:
        input_data (dict): Input data organized by molecule families, where each
            entry contains conformer information with initial and final structures
        calculator: ASE calculator object (e.g., FAIRChemCalculator) to use for energy/force calculations

    Returns:
        dict: Results organized in the following form -
        {
            "molecule_family_1": {
                "conformer_id_1": {
                    "initial": {
                        "atoms": MSONAtoms dictionary of initial structure,
                        "energy": Initial total energy (eV),
                        "forces": Initial forces as a list (eV/Å),
                    },
                    "final": {
                        "atoms": MSONAtoms dictionary of optimized structure,
                        "energy": Final total energy (eV),
                        "forces": Final forces as a list (eV/Å),
                    },
                },
                "conformer_id_2": { ... },
                ...
            },
            "molecule_family_2": { ... },
            ...
        }
    """
    all_results = {}
    for molecule_family in tqdm(input_data):
        conformer_results = {}
        conformers = input_data[molecule_family]
        for conformer in conformers:
            sid = conformer["sid"]
            initial_atoms = conformer["initial_atoms"]
            results = relax_job(initial_atoms, calculator, EVAL_OPT_PARAMETERS)
            conformer_results[sid] = results

        all_results[molecule_family] = conformer_results
    return all_results


def protonation(input_data: dict[str, Any], calculator: Calculator) -> dict[str, Any]:
    """
    Calculate protonation state energies and geometries.

    This function calculates the energies and geometries of different protonation
    states of molecules.

    Args:
        input_data (dict): Input data organized by molecule families, where each
            entry contains different protonation states with initial and final structures
        calculator: ASE calculator object (e.g., FAIRChemCalculator) to use for energy/force calculations

    Returns:
        dict: Results organized in the following form -
        {
            "molecule_family_1": {
                "protonation_state_1": {
                    "initial": {
                        "atoms": MSONAtoms dictionary of initial structure,
                        "energy": Initial total energy (eV),
                        "forces": Initial forces as a list (eV/Å),
                    },
                    "final": {
                        "atoms": MSONAtoms dictionary of optimized structure,
                        "energy": Final total energy (eV),
                        "forces": Final forces as a list (eV/Å),
                    },
                },
                "protonation_state_2": { ... },
                ...
            },
            "molecule_family_2": { ... },
            ...
        }
    """
    all_results = {}
    for molecule_family in tqdm(input_data):
        state_results = {}
        states = input_data[molecule_family]
        for state in states:
            initial_atoms = states[state]["initial_atoms"]
            results = relax_job(initial_atoms, calculator, EVAL_OPT_PARAMETERS)

            state_results[state] = results

        all_results[molecule_family] = state_results
    return all_results


def ieea(input_data: dict[str, Any], calculator: Calculator) -> dict[str, Any]:
    """
    Calculate unoptimized ionization energies and electron affinities.

    This function performs single-point calculations on structures at different
    charge states to evaluate ionization energies (IE) and electron affinities (EA).
    No geometry optimization is performed, testing the MLIP's ability to predict
    energetics of charged species at fixed geometries.

    Args:
        input_data (dict): Input data organized by system identifier, with each
            entry containing structures at different charge and spin states
        calculator: ASE calculator object (e.g., FAIRChemCalculator) to use for energy/force calculations

    Returns:
        dict: Results organized in the following form -
        {
            "identifier_1": {
                "charge_state_1": {
                    "spin_state_1": {
                        "atoms": MSONAtoms dictionary of the structure,
                        "energy": Total energy (eV),
                        "forces": Forces as a list (eV/Å),
                    },
                    "spin_state_2": { ... },
                    ...
                },
                "charge_state_2": { ... },
                ...
            },
            "identifier_2": { ... },
        }
    """
    all_results = {}
    for identifier in tqdm(input_data):
        molecule_results = defaultdict(dict)
        for charge in input_data[identifier]:
            for spin, entry in input_data[identifier][charge].items():
                atoms = entry["atoms"]
                results = single_point_job(atoms, calculator)

                molecule_results[charge][spin] = results

        all_results[identifier] = molecule_results
    return all_results


def spin_gap(input_data: dict[str, Any], calculator: Calculator) -> dict[str, Any]:
    """
    Calculate unoptimized spin gap energies.

    This function performs single-point calculations on structures at different
    spin states to evaluate spin gaps (energy differences between different
    spin multiplicities). No geometry optimization is performed.

    Args:
        input_data (dict): Input data organized by system identifier, with each
            entry containing structures at different spin states
        calculator: ASE calculator object (e.g., FAIRChemCalculator) to use for energy/force calculations

    Returns:
        dict: Results organized in the following form -
        {
            "identifier_1": {
                "spin_state_1": {
                    "atoms": MSONAtoms dictionary of the structure,
                    "energy": Total energy (eV),
                    "forces": Forces as a list (eV/Å),
                },
                "spin_state_2": { ... },
            },
            "identifier_2": { ... },
        }
    """
    all_results = {}
    for identifier in tqdm(input_data):
        molecule_results = {}
        for spin, entry in input_data[identifier].items():
            atoms = entry["atoms"]
            results = single_point_job(atoms, calculator)
            molecule_results[spin] = results

        all_results[identifier] = molecule_results
    return all_results


def ligand_pocket(input_data: dict[str, Any], calculator: Calculator) -> dict[str, Any]:
    """
    Calculate protein-ligand interaction energies and forces.

    This function performs single-point calculations on protein-ligand systems,
    calculating energies and forces for the complex and individual components
    (ligand, pocket, ligand_pocket). This enables evaluation of interaction
    energies and binding affinity predictions.

    Args:
        input_data (dict): Input data organized by system identifier, with each
            entry containing ASE Atoms objects for ligand, pocket, and complex
        calculator: ASE calculator object (e.g., FAIRChemCalculator) to use for energy/force calculations

    Returns:
        dict: Results organized in the following form -
        {
            "identifier_1": {
                "ligand": {
                    "atoms": MSONAtoms dictionary of the structure,
                    "energy": Total energy (eV),
                    "forces": Forces as a list (eV/Å),
                },
                "pocket": {
                    "atoms": MSONAtoms dictionary of the structure,
                    "energy": Total energy (eV),
                    "forces": Forces as a list (eV/Å),
                },
                "ligand_pocket": {
                    "atoms": MSONAtoms dictionary of the structure,
                    "energy": Total energy (eV),
                    "forces": Forces as a list (eV/Å),
                },
            },
            "identifier_2": { ... },
        }
    """
    all_results = {}
    for identifier, entry in tqdm(input_data.items(), total=len(input_data)):
        complex_results = {}
        for mol_type in ["ligand", "pocket", "ligand_pocket"]:
            atoms = entry[mol_type]
            results = single_point_job(atoms, calculator)
            complex_results[mol_type] = results

        all_results[identifier] = complex_results
    return all_results


def ligand_strain(input_data: dict[str, Any], calculator: Calculator) -> dict[str, Any]:
    """
    Calculate ligand strain energies in protein-bound conformations.

    This function calculates strain energies by comparing the energy of a ligand
    in its bioactive (protein-bound) conformation with its global minimum energy
    conformation in the gas phase.


    Args:
        input_data (dict): Input data organized by system identifier, with each
            entry containing:
            - bioactive_conf: Ligand in bioactive conformation
            - conformers: List of (initial, final) conformer pairs for gas phase
        calculator: ASE calculator object (e.g., FAIRChemCalculator) to use for energy/force calculations

    Returns:
        dict: Results organized in the following form -
        {
            "identifier_1": {
                "bioactive": {
                    "atoms": MSONAtoms dictionary of bioactive conformation,
                    "energy": Total energy (eV),
                    "forces": Forces as a list (eV/Å),
                },
                "gas_phase": {
                    "0": {  # conformer index
                        "initial": {
                            "atoms": MSONAtoms dictionary of initial structure,
                            "energy": Initial total energy (eV),
                            "forces": Initial forces as a list (eV/Å),
                        },
                        "final": {
                            "atoms": MSONAtoms dictionary of optimized structure,
                            "energy": Final total energy (eV),
                            "forces": Final forces as a list (eV/Å),
                        }
                    },
                    "1": { ... },
                    ...
                },
            },
            "identifier_2": { ...
            }
        }
    """
    all_results = {}
    for identifier, ligand_system in tqdm(input_data.items()):
        complex_results = {}
        # Bioactive part
        bioactive = ligand_system["bioactive_conf"]
        results = single_point_job(bioactive, calculator)

        complex_results["bioactive"] = results

        # Gas-phase conformers parts
        conformer_prediction = {}
        for idx, initial_atoms in enumerate(ligand_system["conformers"]):
            results = relax_job(initial_atoms, calculator, EVAL_OPT_PARAMETERS)

            conformer_prediction[idx] = results
        complex_results["gas_phase"] = conformer_prediction

        all_results[identifier] = complex_results
    return all_results


def distance_scaling(
    input_data: dict[str, Any], calculator: Calculator
) -> dict[str, Any]:
    """
    Calculate energies and forces at different inter-molecular distances.

    This function performs single-point calculations on molecular systems where
    inter-molecular distances have been systematically varied. This tests the
    MLIP's ability to capture both short-range repulsion and long-range attraction
    in potential energy surfaces.

    Args:
        input_data (dict): Input data organized by domain type (vertical), then
            by system identifier, then by distance scale factor, containing
            ASE Atoms objects at different inter-molecular separations
        calculator: ASE calculator object (e.g., FAIRChemCalculator) to use for energy/force calculations

    Returns:
        dict: Results organized in the following form -
        {
            "vertical_1": {
                "identifier_1": {
                    "short_range_scaled_complex_X": {
                        "atoms": MSONAtoms dictionary of the structure,
                        "energy": Total energy (eV),
                        "forces": Forces as a list (eV/Å),
                    },
                    "short_range_scaled_complex_Y": { ... },
                    "long_range_scaled_complex_Z": { ... },
                    ...
                },
                "identifier_2": { ... },
                ...
            },
            "vertical_2": { ... },
            ...
    """
    all_results = {}
    for vertical, systems in input_data.items():
        species_results = defaultdict(dict)
        for identifier, structures in tqdm(systems.items()):
            for scale, scale_structure in structures.items():
                results = single_point_job(scale_structure, calculator)
                species_results[identifier][scale] = results

        all_results[vertical] = species_results
    return all_results


def singlepoint(input_data: dict[str, Any], calculator: Calculator) -> dict[str, Any]:
    """
    Perform general single-point energy and force calculations.

    This is a general-purpose function for performing single-point calculations
    on arbitrary molecular structures.

    Args:
        input_data (dict): Input data organized by system identifier, with each
            entry containing an ASE Atoms object
        calculator: ASE calculator object (e.g., FAIRChemCalculator) to use for energy/force calculations

    Returns:
        dict: Results organized in the following form -
        {
            "identifier_1": {
                "atoms": MSONAtoms dictionary of the structure,
                "energy": Total energy (eV),
                "forces": Forces as a list (eV/Å),
            },
            "identifier_2": { ... },
        }
    """
    all_results = {}
    for identifier, atoms in tqdm(input_data.items(), total=len(input_data)):
        results = single_point_job(atoms, calculator)
        all_results[identifier] = results
    return all_results



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import copy
import logging
import math
import os
import random
import sys
from collections import defaultdict
from contextlib import nullcontext
from functools import wraps
from typing import TYPE_CHECKING, Protocol

import hydra
import numpy as np
import ray
import torch
import torch.distributed as dist
from ray import remote
from ray.util.scheduling_strategies import PlacementGroupSchedulingStrategy
from torch.distributed.elastic.utils.distributed import get_free_port
from torchtnt.framework import PredictUnit, State

from fairchem.core.common import gp_utils
from fairchem.core.common.distutils import (
    CURRENT_DEVICE_TYPE_STR,
    assign_device_for_local_rank,
    get_device_for_local_rank,
    setup_env_local_multi_gpu,
)
from fairchem.core.datasets.atomic_data import AtomicData
from fairchem.core.units.mlip_unit import InferenceSettings
from fairchem.core.units.mlip_unit.utils import (
    get_backbone_class_from_checkpoint,
    load_inference_model,
    tf32_context_manager,
)

if TYPE_CHECKING:
    from ase import Atoms

    from fairchem.core.units.mlip_unit.api.inference import MLIPInferenceCheckpoint


def collate_predictions(predict_fn):
    @wraps(predict_fn)
    def collated_predict(
        predict_unit, data: AtomicData, undo_element_references: bool = True
    ):
        # Get the full prediction dictionary from the original predict method
        preds = predict_fn(predict_unit, data, undo_element_references)
        collated_preds = defaultdict(list)
        for i, dataset in enumerate(data.dataset):
            for task in predict_unit.dataset_to_tasks[dataset]:
                if task.level == "system":
                    collated_preds[task.property].append(
                        preds[task.name][i].unsqueeze(0)
                    )
                elif task.level == "atom":
                    collated_preds[task.property].append(
                        preds[task.name][data.batch == i]
                    )
                else:
                    raise RuntimeError(
                        f"Unrecognized task level={task.level} found in data batch at position {i}"
                    )

        return {prop: torch.cat(val) for prop, val in collated_preds.items()}

    return collated_predict


class MLIPPredictUnitProtocol(Protocol):
    def predict(self, data: AtomicData, undo_element_references: bool) -> dict: ...

    def validate_atoms_data(self, atoms: Atoms, task_name: str) -> None: ...

    @property
    def dataset_to_tasks(self) -> dict[str, list]: ...


def merge_uma_model(model, data):
    logging.info("Calling merge_uma_model")
    # merge the backbone
    model.backbone = model.backbone.merge_MOLE_model(data)

    # merge any heads
    new_output_heads = torch.nn.ModuleDict()
    for head_name, head in model.output_heads.items():
        if hasattr(head, "merge_MOLE_model"):
            new_output_heads[head_name] = head.merge_MOLE_model(data)
        else:
            new_output_heads[head_name] = head
    model.output_heads = new_output_heads


class MLIPPredictUnit(PredictUnit[AtomicData], MLIPPredictUnitProtocol):
    def __init__(
        self,
        inference_model_path: str,
        device: str = "cpu",
        overrides: dict | None = None,
        inference_settings: InferenceSettings | None = None,
        seed: int = 41,
        atom_refs: dict | None = None,
        form_elem_refs: dict | None = None,
        assert_on_nans: bool = False,
    ):
        super().__init__()
        os.environ[CURRENT_DEVICE_TYPE_STR] = device

        self.set_seed(seed)
        self._setup_refs(atom_refs, form_elem_refs)

        if inference_settings is None:
            inference_settings = InferenceSettings()
        self._setup_threads(inference_settings)

        if inference_settings.wigner_cuda:
            logging.warning(
                "The wigner_cuda flag is deprecated and will be removed in future versions."
            )

        # Load checkpoint first to get model type
        checkpoint = torch.load(
            inference_model_path, map_location="cpu", weights_only=False
        )

        # Build model-specific overrides
        final_overrides = self._build_overrides_from_settings(
            checkpoint, overrides, inference_settings
        )

        # Load model with overrides, passing pre-loaded checkpoint
        self.model, checkpoint = load_inference_model(
            inference_model_path,
            use_ema=True,
            overrides=final_overrides,
            preloaded_checkpoint=checkpoint,
        )

        # Model sets up tasks
        self.model.module.setup_tasks(checkpoint.tasks_config)

        self._setup_device(device)

        self.model.eval()
        self.lazy_model_intialized = False
        self.inference_settings = inference_settings
        self.assert_on_nans = assert_on_nans

        if self.model.module.direct_forces:
            logging.warning(
                "This is a direct-force model. Direct force predictions may lead to "
                "discontinuities in the potential energy surface and energy conservation errors."
            )

    @property
    def direct_forces(self) -> bool:
        return self.model.module.direct_forces

    @property
    def dataset_to_tasks(self) -> dict[str, list]:
        return self.model.module.dataset_to_tasks

    @property
    def tasks(self) -> dict:
        return self.model.module.tasks

    def set_seed(self, seed: int) -> None:
        """
        Initialize random seeds.
        """
        logging.debug(f"Setting random seed to {seed}")
        self._seed = seed
        random.seed(seed)
        np.random.seed(seed)
        torch.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)

    def _setup_refs(self, atom_refs: dict | None, form_elem_refs: dict | None) -> None:
        """
        Setup element references.
        """
        self.atom_refs = (
            {task.replace("_elem_refs", ""): refs for task, refs in atom_refs.items()}
            if atom_refs
            else {}
        )
        self.form_elem_refs = form_elem_refs or {}

    def _setup_threads(self, settings: InferenceSettings) -> None:
        """
        Configure thread settings.
        """
        if settings.torch_num_threads is not None:
            torch.set_num_threads(settings.torch_num_threads)
            torch.set_num_interop_threads(settings.torch_num_threads)

    def _setup_device(self, device: str) -> None:
        """
        Setup inference device.
        """
        assert device in ["cpu", "cuda"], "device must be either 'cpu' or 'cuda'"
        self.device = get_device_for_local_rank() if device == "cuda" else "cpu"

    def _build_overrides_from_settings(
        self,
        checkpoint: MLIPInferenceCheckpoint,
        user: dict | None,
        settings: InferenceSettings,
    ) -> dict:
        """Build backbone config overrides by delegating to model-specific logic."""
        overrides = {} if user is None else dict(user)
        if "backbone" not in overrides:
            overrides["backbone"] = {}

        # Delegate to model-specific classmethod
        backbone_cls = get_backbone_class_from_checkpoint(checkpoint)
        backbone_overrides = backbone_cls.build_inference_settings(settings)

        overrides["backbone"].update(backbone_overrides)

        # User overrides take precedence
        if user is not None and "backbone" in user:
            overrides["backbone"].update(user["backbone"])

        return overrides

    def move_to_device(self):
        self.model.to(self.device)
        for task in self.model.module.tasks.values():
            task.normalizer.to(self.device)
            if task.element_references is not None:
                task.element_references.to(self.device)

    def validate_atoms_data(self, atoms: Atoms, task_name: str) -> None:
        """
        Validate and set defaults for calculator input data.

        Delegates to the model's backbone for model-specific validation.
        """
        self.model.module.validate_atoms_data(atoms, task_name)

    def predict_step(self, state: State, data: AtomicData) -> dict[str, torch.tensor]:
        return self.predict(data)

    @collate_predictions
    def predict(
        self, data: AtomicData, undo_element_references: bool = True
    ) -> dict[str, torch.tensor]:
        if not self.lazy_model_intialized:
            self._lazy_init(data)

        data_device = data.to(self.device).clone()

        # Model handles any per-prediction checks (e.g., MOLE consistency)
        self.model.module.on_predict_check(data_device)

        return self._run_inference(data_device, undo_element_references)

    def _lazy_init(self, data: AtomicData) -> None:
        """
        Lazy initialization on first predict call.
        """
        # Model handles its own preparation (MOLE merge, eval mode, etc.)
        self.model.module.prepare_for_inference(data, self.inference_settings)

        self.move_to_device()

        if self.inference_settings.compile:
            logging.warning(
                "Model is being compiled this might take a while for the first time"
            )
            self.model = torch.compile(self.model, dynamic=True)

        self.lazy_model_intialized = True

    def _run_inference(self, data: AtomicData, undo_refs: bool) -> dict:
        """
        Execute model inference.
        """
        inference_context = torch.no_grad() if self.direct_forces else nullcontext()
        tf32_context = (
            tf32_context_manager() if self.inference_settings.tf32 else nullcontext()
        )

        with inference_context, tf32_context:
            output = self.model(data)
            return self._process_outputs(data, output, undo_refs)

    def _process_outputs(self, data: AtomicData, output: dict, undo_refs: bool) -> dict:
        """
        Denormalize and post-process model outputs.
        """
        pred_output = {}
        for task_name, task in self.model.module.tasks.items():
            pred_output[task_name] = task.normalizer.denorm(
                output[task_name][task.property]
            )
            if self.assert_on_nans:
                assert (
                    torch.isfinite(pred_output[task_name]).all()
                ), f"NaNs/Infs found in prediction for task {task_name}.{task.property}"
            if undo_refs and task.element_references is not None:
                pred_output[task_name] = task.element_references.undo_refs(
                    data, pred_output[task_name]
                )
        return pred_output


def move_tensors_to_cpu(data):
    """
    Recursively move all PyTorch tensors in a nested data structure to CPU.

    Args:
        data: Input data structure (dict, list, tuple, tensor, or other)

    Returns:
        Data structure with all tensors moved to CPU
    """
    if isinstance(data, torch.Tensor):
        return data.cpu()
    elif isinstance(data, dict):
        return {key: move_tensors_to_cpu(value) for key, value in data.items()}
    elif isinstance(data, list):
        return [move_tensors_to_cpu(item) for item in data]
    elif isinstance(data, tuple):
        return tuple(move_tensors_to_cpu(item) for item in data)
    else:
        # Return as-is for non-tensor types (str, int, float, etc.)
        return data


class MLIPWorkerLocal:
    def __init__(
        self,
        worker_id: int,
        world_size: int,
        predictor_config: dict,
        master_port: int | None = None,
        master_address: str | None = None,
    ):
        self.worker_id = worker_id
        self.world_size = world_size
        self.predictor_config = predictor_config
        self.master_address = (
            ray.util.get_node_ip_address() if master_address is None else master_address
        )
        self.master_port = get_free_port() if master_port is None else master_port
        self.is_setup = False
        self.last_received_atomic_data = None

    def get_master_address_and_port(self):
        return (self.master_address, self.master_port)

    def get_device_for_local_rank(self):
        return get_device_for_local_rank()

    def _distributed_setup(
        self,
    ):
        # initialize distributed environment
        # TODO, this wont work for multi-node, need to fix master addr
        logging.info(f"Initializing worker {self.worker_id}...")
        setup_env_local_multi_gpu(self.worker_id, self.master_port, self.master_address)

        device = self.predictor_config.get("device", "cpu")
        assign_device_for_local_rank(device == "cpu", 0)
        backend = "gloo" if device == "cpu" else "nccl"
        dist.init_process_group(
            backend=backend,
            rank=self.worker_id,
            world_size=self.world_size,
        )
        gp_utils.setup_graph_parallel_groups(self.world_size, backend)
        self.predict_unit = hydra.utils.instantiate(self.predictor_config)
        self.device = get_device_for_local_rank()
        logging.info(
            f"Worker {self.worker_id}, gpu_id: {ray.get_gpu_ids()}, loaded predict unit: {self.predict_unit}, "
            f"on port {self.master_port}, with device: {self.device}, config: {self.predictor_config}"
        )
        self.is_setup = True

    def predict(
        self, data: AtomicData, use_nccl: bool = False
    ) -> dict[str, torch.tensor] | None:
        if not self.is_setup:
            self._distributed_setup()

        out = self.predict_unit.predict(data)
        if self.worker_id == 0:
            return move_tensors_to_cpu(out)

        if self.worker_id != 0 and use_nccl:
            self.last_received_atomic_data = data.to(self.device)
            while True:
                torch.distributed.broadcast(self.last_received_atomic_data.pos, src=0)
                torch.distributed.broadcast(self.last_received_atomic_data.cell, src=0)
                self.predict_unit.predict(self.last_received_atomic_data)

        return None


@remote
class MLIPWorker(MLIPWorkerLocal):
    pass


class ParallelMLIPPredictUnit(MLIPPredictUnitProtocol):
    def __init__(
        self,
        inference_model_path: str,
        device: str = "cpu",
        overrides: dict | None = None,
        inference_settings: InferenceSettings | None = None,
        seed: int = 41,
        atom_refs: dict | None = None,
        form_elem_refs: dict | None = None,
        assert_on_nans: bool = False,
        num_workers: int = 1,
        num_workers_per_node: int = 8,
        log_level: int = logging.INFO,
    ):
        super().__init__()
        _mlip_pred_unit = MLIPPredictUnit(
            inference_model_path=inference_model_path,
            device="cpu",
            overrides=overrides,
            inference_settings=inference_settings,
            seed=seed,
            atom_refs=atom_refs,
            form_elem_refs=form_elem_refs,
        )
        if inference_settings is None:
            inference_settings = InferenceSettings()
        self.inference_settings = inference_settings
        self._dataset_to_tasks = copy.deepcopy(_mlip_pred_unit.dataset_to_tasks)
        self._validate_atoms_data_fn = _mlip_pred_unit.model.module.validate_atoms_data

        predict_unit_config = {
            "_target_": "fairchem.core.units.mlip_unit.predict.MLIPPredictUnit",
            "inference_model_path": inference_model_path,
            "device": device,
            "overrides": overrides,
            "inference_settings": inference_settings,
            "seed": seed,
            "atom_refs": atom_refs,
            "form_elem_refs": form_elem_refs,
            "assert_on_nans": assert_on_nans,
        }

        logging.basicConfig(
            level=log_level,
            force=True,
            stream=sys.stdout,
            format="%(asctime)s %(levelname)s [%(processName)s] %(name)s: %(message)s",
        )
        # Optional: keep Ray/uvicorn chatty logs in check
        logging.getLogger("ray").setLevel(log_level)
        logging.getLogger("uvicorn").setLevel(log_level)
        if not ray.is_initialized():
            # in CI envrionment, we want to control the number of CPUs allocated to limit the pool of IDLE ray workers
            if os.environ.get("CI"):
                logging.info(
                    f"CI environment detected, initializing ray with limited CPUs: {num_workers_per_node}"
                )
                ray.init(
                    logging_level=log_level,
                    num_cpus=num_workers_per_node,
                    runtime_env={
                        "env_vars": {"RAY_DEBUG": "1"},
                    },
                )
            else:
                ray.init(logging_level=log_level)

        self.atomic_data_on_device = None

        num_nodes = math.ceil(num_workers / num_workers_per_node)
        num_workers_on_node_array = [num_workers_per_node] * num_nodes
        if num_workers % num_workers_per_node > 0:
            num_workers_on_node_array[-1] = num_workers % num_workers_per_node
        logging.info(
            f"Creating placement groups with {num_workers_on_node_array} workers on {device}"
        )

        # first create one placement group for each node
        num_gpu_per_worker = 1 if device == "cuda" else 0
        placement_groups = []
        for workers in num_workers_on_node_array:
            bundle = {"CPU": workers}
            if device == "cuda":
                bundle["GPU"] = workers
            pg = ray.util.placement_group([bundle], strategy="STRICT_PACK")
            placement_groups.append(pg)
        ray.get(pg.ready())  # Wait for each placement group to be scheduled

        # Need to still place worker to occupy space, otherwise ray double books this GPU
        rank0_worker = MLIPWorker.options(
            num_gpus=num_gpu_per_worker,
            scheduling_strategy=PlacementGroupSchedulingStrategy(
                placement_group=placement_groups[0],
                placement_group_bundle_index=0,  # Use the first (and only) bundle in the PG
                placement_group_capture_child_tasks=True,  # Ensure child tasks also run in this PG
            ),
        ).remote(0, num_workers, predict_unit_config)

        local_gpu_or_cpu = ray.get(rank0_worker.get_device_for_local_rank.remote())
        os.environ[CURRENT_DEVICE_TYPE_STR] = local_gpu_or_cpu

        self.workers = []
        self.local_rank0 = MLIPWorkerLocal(
            worker_id=0,
            world_size=num_workers,
            predictor_config=predict_unit_config,
        )
        master_addr, master_port = self.local_rank0.get_master_address_and_port()
        logging.info(f"Started rank0 on {master_addr}:{master_port}")

        # next place all ranks in order and pack them on placement groups
        # ie: rank0-7 -> placement group 0, 8->15 -> placement group 1 etc.
        worker_id = 0
        for pg_idx, pg in enumerate(placement_groups):
            workers = num_workers_on_node_array[pg_idx]
            logging.info(
                f"Launching workers for placement group {pg_idx} (Node {pg_idx}), workers={workers}"
            )

            for i in range(workers):
                # skip the first one because it's already been initialized above
                if pg_idx == 0 and i == 0:
                    worker_id += 1
                    continue
                # Each actor requests 1 worker worth of resources and uses the specific placement group
                actor = MLIPWorker.options(
                    num_gpus=num_gpu_per_worker,
                    scheduling_strategy=PlacementGroupSchedulingStrategy(
                        placement_group=pg,
                        placement_group_bundle_index=0,  # Use the first (and only) bundle in the PG
                        placement_group_capture_child_tasks=True,  # Ensure child tasks also run in this PG
                    ),
                ).remote(
                    worker_id,
                    num_workers,
                    predict_unit_config,
                    master_port,
                    master_addr,
                )
                self.workers.append(actor)
                worker_id += 1

    def predict(self, data: AtomicData) -> dict[str, torch.tensor]:
        # put the reference in the object store only once
        if not self.inference_settings.merge_mole or self.atomic_data_on_device is None:
            data_ref = ray.put(data)
            # this will put the ray works into an infinite loop listening for broadcasts
            _futures = [
                w.predict.remote(data_ref, use_nccl=self.inference_settings.merge_mole)
                for w in self.workers
            ]
            self.atomic_data_on_device = data.clone()
        else:
            self.atomic_data_on_device.pos = data.pos.to(self.local_rank0.device)
            self.atomic_data_on_device.cell = data.cell.to(self.local_rank0.device)
            torch.distributed.broadcast(self.atomic_data_on_device.pos, src=0)
            torch.distributed.broadcast(self.atomic_data_on_device.cell, src=0)

        return self.local_rank0.predict(self.atomic_data_on_device)

    def validate_atoms_data(self, atoms: Atoms, task_name: str) -> None:
        """
        Validate and set defaults for calculator input data.

        Delegates to the model's validate_atoms_data captured at init time.
        """
        self._validate_atoms_data_fn(atoms, task_name)

    @property
    def dataset_to_tasks(self) -> dict[str, list]:
        return self._dataset_to_tasks


class BatchServerPredictUnit(MLIPPredictUnitProtocol):
    """
    PredictUnit wrapper that uses Ray Serve for batched inference.

    This provides a clean interface compatible with MLIPPredictUnitProtocol
    while leveraging Ray Serve's batching capabilities under the hood.
    """

    def __init__(
        self,
        server_handle,
        predict_unit: MLIPPredictUnit,
    ):
        """
        Args:
            server_handle: Ray Serve deployment handle for BatchPredictServer
            predict_unit: Local MLIPPredictUnit used for input validation.
                Validation must run locally because it mutates atoms.info.
        """
        self.server_handle = server_handle
        self._predict_unit = predict_unit

    def predict(self, data: AtomicData, undo_element_references: bool = True) -> dict:
        """
        Args:
            data: AtomicData object (single system)
            undo_element_references: Whether to undo element references

        Returns:
            Prediction dictionary
        """
        result = self.server_handle.predict.remote(
            data, undo_element_references
        ).result()
        return result

    def validate_atoms_data(self, atoms: Atoms, task_name: str) -> None:
        """
        Validate and set defaults for calculator input data.

        Runs locally (not via Ray Serve) because validation mutates atoms.info.
        """
        self._predict_unit.validate_atoms_data(atoms, task_name)

    @property
    def dataset_to_tasks(self) -> dict:
        return self.server_handle.get_predict_unit_attribute.remote(
            "dataset_to_tasks"
        ).result()

    @property
    def atom_refs(self) -> dict | None:
        return self.server_handle.get_predict_unit_attribute.remote(
            "atom_refs"
        ).result()

    @property
    def inference_settings(self) -> InferenceSettings:
        return self.server_handle.get_predict_unit_attribute.remote(
            "inference_settings"
        ).result()



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import heapq
import logging
from typing import TYPE_CHECKING, Any, Literal

import numba
import numpy as np
import torch
import torch.distributed
from torch.utils.data import BatchSampler, Dataset, DistributedSampler
from typing_extensions import deprecated, override

from fairchem.core.common import distutils, gp_utils
from fairchem.core.datasets import data_list_collater
from fairchem.core.datasets.base_dataset import (
    UnsupportedDatasetError,
)

if TYPE_CHECKING:
    from numpy.typing import NDArray

    from fairchem.core.datasets.atomic_data import AtomicData


@deprecated(
    "OCPColatter is deprecated. Please use data_list_collater optionally with functools.partial to set defaults"
)
class OCPCollater:
    def __init__(self, otf_graph: bool = False) -> None:
        self.otf_graph = otf_graph

    def __call__(self, data_list: list[AtomicData]) -> AtomicData:
        return data_list_collater(data_list, otf_graph=self.otf_graph)


@numba.njit
def _balanced_partition(sizes: NDArray[np.int_], num_parts: int):
    """
    Greedily partition the given set by always inserting
    the largest element into the smallest partition.
    """
    sort_idx = np.argsort(-sizes)  # Sort in descending order
    heap = [(sizes[idx], [idx]) for idx in sort_idx[:num_parts]]
    heapq.heapify(heap)
    for idx in sort_idx[num_parts:]:
        smallest_part = heapq.heappop(heap)
        new_size = smallest_part[0] + sizes[idx]
        new_idx = smallest_part[1] + [
            idx
        ]  # TODO should this be append to save time/space
        heapq.heappush(heap, (new_size, new_idx))
    return [part[1] for part in heap]


class StatefulDistributedSampler(DistributedSampler):
    """
    More fine-grained state DataSampler that uses training iteration and epoch
    both for shuffling data. PyTorch DistributedSampler only uses epoch
    for the shuffling and starts sampling data from the start. In case of training
    on very large data, we train for one epoch only and when we resume training,
    we want to resume the data sampler from the training iteration.
    """

    def __init__(self, dataset, batch_size, **kwargs):
        """
        Initializes the instance of StatefulDistributedSampler. Random seed is set
        for the epoch set and data is shuffled. For starting the sampling, use
        the start_iter (set to 0 or set by checkpointing resuming) to
        sample data from the remaining images.

        Args:
            dataset (Dataset): Pytorch dataset that sampler will shuffle
            batch_size (int): batch size we want the sampler to sample
            seed (int): Seed for the torch generator.
        """
        super().__init__(dataset=dataset, **kwargs)

        self.start_iter = 0
        self.batch_size = batch_size
        assert self.batch_size > 0, "batch_size not set for the sampler"
        logging.info(f"rank: {self.rank}: Sampler created...")

    def __iter__(self):
        # TODO: For very large datasets, even virtual datasets this might slow down
        # or not work correctly. The issue is that we enumerate the full list of all
        # samples in a single epoch, and manipulate this list directly. A better way
        # of doing this would be to keep this sequence strictly as an iterator
        # that stores the current state (instead of the full sequence)
        distributed_sampler_sequence = super().__iter__()
        if self.start_iter > 0:
            for i, _ in enumerate(distributed_sampler_sequence):
                if i == self.start_iter * self.batch_size - 1:
                    break
        return distributed_sampler_sequence

    def set_epoch_and_start_iteration(self, epoch, start_iter):
        self.set_epoch(epoch)
        self.start_iter = start_iter


def _ensure_supported(dataset: Any):
    if not isinstance(dataset, Dataset):
        raise UnsupportedDatasetError("BalancedBatchSampler requires a dataset.")

    if not dataset.metadata_hasattr("natoms"):
        raise UnsupportedDatasetError(
            "BalancedBatchSampler requires a dataset that has a metadata attributed with number of atoms."
        )

    logging.debug(f"BalancedBatchSampler: Resolved dataset to {type(dataset)}")
    return dataset


class BalancedBatchSampler(BatchSampler):
    def __init__(
        self,
        dataset: Dataset,
        *,
        batch_size: int,
        num_replicas: int,
        rank: int,
        device: (
            torch.device | None
        ) = None,  # deprecated, unused variable for backwards compat
        seed: int,
        mode: bool | Literal["atoms"] = "atoms",
        shuffle: bool = True,
        on_error: Literal["warn_and_balance", "warn_and_no_balance", "raise"] = "raise",
        drop_last: bool = False,
    ):
        """
        Initializes a BalancedBatchSampler object.

        Args:
            dataset (Dataset): The dataset to sample from.
            batch_size (int): The size of each batch.
            num_replicas (int): The number of processes participating in distributed training.
            rank (int): The rank of the current process in distributed training.
            device (torch.device): The device to use for the batches.
            mode (str or bool, optional): The mode to use for balancing the batches. Defaults to "atoms".
            shuffle (bool, optional): Whether to shuffle the samples. Defaults to True.
            on_error (Literal["warn_and_balance", "warn_and_no_balance", "raise"], optional): The action to take when an error occurs (i.e., when we have an invalid dataset). Defaults to "raise".
                - "warn_and_balance": Raise a warning and balance the batch by manually loading the data samples and counting the number of nodes (this is slow).
                - "warn_and_no_balance": Raise a warning and do not do any balancing.
                - "raise": Raise an error.
            drop_last (bool, optional): Whether to drop the last incomplete batch. Defaults to False.
        """
        self.disabled = False
        self.on_error = on_error

        if mode is False:
            logging.warning(f"Disabled BalancedBatchSampler because {mode=}.")
            self.disabled = True
        elif mode.lower() != "atoms":
            raise ValueError(
                f"Only mode='atoms' or mode=True is supported, got {mode=}."
            )
        elif num_replicas == 1:
            logging.warning(f"Disabled BalancedBatchSampler because {num_replicas=}.")
            self.disabled = True

        try:
            dataset = _ensure_supported(dataset)
        except UnsupportedDatasetError as error:
            if self.on_error == "raise":
                raise error
            if self.on_error == "warn_and_balance":
                logging.warning(
                    f"Failed to get data sizes from metadata, loading data to get sizes (THIS IS SLOW). {error}"
                )
            elif self.on_error == "warn_and_no_balance":
                logging.warning(
                    f"Failed to get data sizes, falling back to uniform partitioning. {error}"
                )
            else:
                raise ValueError(f"Unknown on_error={self.on_error}") from error

        sampler = StatefulDistributedSampler(
            dataset,
            num_replicas=num_replicas,
            rank=rank,
            shuffle=shuffle,
            drop_last=drop_last,
            batch_size=batch_size,
            seed=seed,
        )

        super().__init__(sampler, batch_size=batch_size, drop_last=drop_last)
        self.device = (
            device if device is not None else distutils.get_device_for_local_rank()
        )

        logging.info(
            f"Created BalancedBatchSampler with {sampler=}, {batch_size=}, {drop_last=}"
        )

    def _get_natoms(self, batch_idx: list[int]):
        if self.sampler.dataset.metadata_hasattr("natoms"):
            return np.array(
                self.sampler.dataset.get_metadata("natoms", batch_idx)
            ).reshape(-1)
        if self.on_error == "warn_and_balance":
            return np.array([self.sampler.dataset[idx].num_nodes for idx in batch_idx])
        return None

    def set_epoch_and_start_iteration(self, epoch: int, start_iteration: int) -> None:
        if not isinstance(self.sampler, StatefulDistributedSampler):
            if start_iteration != 0:
                raise NotImplementedError(
                    f"{type(self.single_sampler)} does not support resuming from a nonzero step."
                )
            self.sampler.set_epoch(epoch)
        else:
            self.sampler.set_epoch_and_start_iteration(epoch, start_iteration)

    def set_epoch(self, epoch: int) -> None:
        if isinstance(self.sampler, DistributedSampler):
            self.sampler.set_epoch(epoch)

    @staticmethod
    def _dist_enabled():
        return torch.distributed.is_available() and torch.distributed.is_initialized()

    @override
    def __iter__(self):
        if self.disabled or not self._dist_enabled():
            yield from super().__iter__()
            return

        for batch_idx in super().__iter__():
            sizes = self._get_natoms(batch_idx)
            if sizes is None:  # on_error == "warn_and_no_balance" is set
                yield batch_idx
                continue

            idx_sizes = torch.stack(
                [
                    torch.tensor(batch_idx, device=self.device),
                    torch.tensor(sizes, device=self.device),
                ]
            )
            idx_sizes_all = distutils.all_gather(idx_sizes, device=self.device)
            idx_sizes_all = torch.cat(idx_sizes_all, dim=-1).cpu()
            if gp_utils.initialized():
                idx_sizes_all = torch.unique(input=idx_sizes_all, dim=1)
            idx_all = idx_sizes_all[0]
            sizes_all = idx_sizes_all[1]

            local_idx_balanced = _balanced_partition(
                sizes_all.numpy(),
                num_parts=self.sampler.num_replicas,
            )
            # Since DistributedSampler pads the last batch
            # this should always have an entry for each replica.
            yield idx_all[local_idx_balanced[self.sampler.rank]]



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

from fairchem.core.calculate._batch import InferenceBatcher
from fairchem.core.calculate.ase_calculator import (
    FAIRChemCalculator,
    FormationEnergyCalculator,
)

__all__ = ["FAIRChemCalculator", "FormationEnergyCalculator", "InferenceBatcher"]



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import os
import traceback
from typing import TYPE_CHECKING, Any, ClassVar

import numpy as np
import pandas as pd
from ase.io.jsonio import encode
from tqdm import tqdm

from fairchem.core.components.calculate._calculate_runner import CalculateRunner
from fairchem.core.components.calculate.recipes.utils import (
    get_property_dict_from_atoms,
)

if TYPE_CHECKING:
    from collections.abc import Sequence

    from ase.calculators.calculator import Calculator

    from fairchem.core.datasets.atoms_sequence import AtomsSequence


class SinglePointRunner(CalculateRunner):
    """Perform a single point calculation of several structures/molecules.

    This class handles the single point calculation of atomic structures using a specified calculator,
    processes the input data in chunks, and saves the results.
    """

    result_glob_pattern: ClassVar[str] = "singlepoint_*-*.json.gz"

    def __init__(
        self,
        calculator: Calculator,
        input_data: AtomsSequence,
        calculate_properties: Sequence[str] = ["energy"],
        normalize_properties_by: dict[str, str] | None = None,
        save_target_properties: Sequence[str] | None = None,
        save_atoms: bool = True,
    ):
        """Initialize the SinglePointRunner.

        Args:
            calculator: ASE calculator to use for energy and force calculations
            input_data: Dataset containing atomic structures to process
            calculate_properties: Sequence of properties to calculate
            normalize_properties_by (dict[str, str] | None): Dictionary mapping property names to natoms or a key in
                atoms.info to normalize by
            save_target_properties (Sequence[str] | None): Sequence of target property names to save in the results file
                These properties need to be available using atoms.get_properties or present in the atoms.info dictionary
            save_atoms (bool): Whether to save json serialized Atoms in the results file.
        """
        self._calculate_properties = calculate_properties
        self._normalize_properties_by = normalize_properties_by or {}
        self._save_target_properties = (
            save_target_properties if save_target_properties is not None else []
        )
        self._save_atoms = save_atoms

        super().__init__(calculator=calculator, input_data=input_data)

    def calculate(self, job_num: int = 0, num_jobs: int = 1) -> list[dict[str, Any]]:
        """Perform singlepoint calculations on a subset of structures.

        Splits the input data into chunks and processes the chunk corresponding to job_num.

        Args:
            job_num (int, optional): Current job number in array job. Defaults to 0.
            num_jobs (int, optional): Total number of jobs in array. Defaults to 1.

        Returns:
            list[dict[str, Any]] - List of dictionaries containing calculation results
        """
        all_results = []
        chunk_indices = np.array_split(range(len(self.input_data)), num_jobs)[job_num]
        for i in tqdm(chunk_indices, desc="Running singlepoint calculations"):
            atoms = self.input_data[i]
            results = {
                "sid": atoms.info.get("sid", i),
                "natoms": len(atoms),
            }

            if self._save_atoms:
                results["atoms"] = encode(atoms)

            # add target properties if requested
            target_properties = get_property_dict_from_atoms(
                self._save_target_properties, atoms, self._normalize_properties_by
            )
            results.update(
                {f"{key}_target": target_properties[key] for key in target_properties}
            )

            try:
                atoms.calc = self.calculator
                results.update(
                    get_property_dict_from_atoms(
                        self._calculate_properties, atoms, self._normalize_properties_by
                    )
                )
                results.update(
                    {
                        "errors": "",
                        "traceback": "",
                    }
                )
            except Exception as ex:  # TODO too broad-figure out which to catch
                results.update(dict.fromkeys(self._calculate_properties, np.nan))
                results.update(
                    {
                        "errors": f"{ex!r}",
                        "traceback": traceback.format_exc(),
                    }
                )

            all_results.append(results)

        return all_results

    def write_results(
        self,
        results: list[dict[str, Any]],
        results_dir: str,
        job_num: int = 0,
        num_jobs: int = 1,
    ) -> None:
        """Write calculation results to a compressed JSON file.

        Args:
            results: List of dictionaries containing energy and forces results
            results_dir: Directory path where results will be saved
            job_num: Index of the current job
            num_jobs: Total number of jobs
        """
        results_df = pd.DataFrame(results)
        results_df.to_json(
            os.path.join(results_dir, f"singlepoint_{num_jobs}-{job_num}.json.gz")
        )

    def save_state(self, checkpoint_location: str, is_preemption: bool = False) -> bool:
        return True



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.

modified from troch_geometric Data class
"""

from __future__ import annotations

import copy
import re
from collections.abc import Sequence
from typing import List, Optional, Union

import ase
import ase.db.sqlite
import numpy as np
import torch
from ase.calculators.singlepoint import SinglePointCalculator, SinglePointDFTCalculator
from ase.constraints import FixAtoms
from ase.geometry import wrap_positions
from ase.stress import full_3x3_to_voigt_6_stress, voigt_6_to_full_3x3_stress
from monty.dev import requires

try:
    from pymatgen.io.ase import AseAtomsAdaptor

    pmg_installed = True
except ImportError:
    AseAtomsAdaptor = None
    pmg_installed = False


IndexType = Union[slice, torch.Tensor, np.ndarray, Sequence]

# these are all currently certainly output by the current a2g
# except for tags, all fields are required for network inference.
_REQUIRED_KEYS = [
    "pos",
    "atomic_numbers",
    "cell",
    "pbc",
    "natoms",
    "edge_index",
    "cell_offsets",
    "nedges",
    "charge",
    "spin",
    "fixed",
    "tags",
]

_OPTIONAL_KEYS = ["energy", "forces", "stress", "dataset"]

# TODO: potential future keys
# ["virials", "atom_attr", "edge_attr"]


def size_repr(key: str, item: torch.Tensor, indent=0) -> str:
    indent_str = " " * indent
    if torch.is_tensor(item) and item.dim() == 0:
        out = item.item()
    elif torch.is_tensor(item):
        out = str(list(item.size()))
    elif isinstance(item, (List, tuple)):
        out = str([len(item)])
    elif isinstance(item, dict):
        lines = [indent_str + size_repr(k, v, 2) for k, v in item.items()]
        out = "{\n" + ",\n".join(lines) + "\n" + indent_str + "}"
    elif isinstance(item, str):
        out = f'"{item}"'
    else:
        out = str(item)

    return f"{indent_str}{key}={out}"


@requires(pmg_installed, message="Requires `pymatgen` to be installed")
def get_neighbors_pymatgen(atoms: ase.Atoms, cutoff, max_neigh):
    """Performs nearest neighbor search and returns edge index, distances,
    and cell offsets"""
    struct = AseAtomsAdaptor.get_structure(atoms)

    # tol of 1e-8 should remove all self loops
    _c_index, _n_index, _offsets, n_distance = struct.get_neighbor_list(
        r=cutoff, numerical_tol=1e-8, exclude_self=True
    )
    _nonmax_idx = []
    for i in range(len(atoms)):
        idx_i = (_c_index == i).nonzero()[0]
        # sort neighbors by distance, remove edges larger than max_neighbors
        idx_sorted = np.argsort(n_distance[idx_i])[:max_neigh]
        _nonmax_idx.append(idx_i[idx_sorted])
    _nonmax_idx = np.concatenate(_nonmax_idx)

    _c_index = _c_index[_nonmax_idx]
    _n_index = _n_index[_nonmax_idx]
    n_distance = n_distance[_nonmax_idx]
    _offsets = _offsets[_nonmax_idx]

    return _c_index, _n_index, n_distance, _offsets


def reshape_features(
    c_index: np.ndarray,
    n_index: np.ndarray,
    n_distance: np.ndarray,
    offsets: np.ndarray,
):
    """Stack center and neighbor index and reshapes distances,
    takes in np.arrays and returns torch tensors"""
    edge_index = torch.LongTensor(np.vstack((n_index, c_index)))
    edge_distances = torch.FloatTensor(n_distance)
    cell_offsets = torch.FloatTensor(offsets)

    # remove distances smaller than a tolerance ~ 0. The small tolerance is
    # needed to correct for pymatgen's neighbor_list returning self atoms
    # in a few edge cases.
    nonzero = torch.where(edge_distances >= 1e-8)[0]
    edge_index = edge_index[:, nonzero]
    cell_offsets = cell_offsets[nonzero]

    return edge_index, cell_offsets


class AtomicData:
    def __init__(
        self,
        pos: torch.Tensor,  # (num_node, 3)
        atomic_numbers: torch.Tensor,  # (num_node,)
        cell: torch.Tensor,  # (num_graph, 3, 3)
        pbc: torch.Tensor,  # (num_graph, 3)
        natoms: torch.Tensor,  # (1,)
        edge_index: torch.Tensor,  # (2, num_edge)
        cell_offsets: torch.Tensor,  # (num_edge, 3)
        nedges: torch.Tensor,  # (1,)
        charge: torch.Tensor,  # (num_graph,)
        spin: torch.Tensor,  # (num_graph,)
        fixed: torch.Tensor,  # (num_node,)
        tags: torch.Tensor,  # (num_node,)
        energy: torch.Tensor | None = None,  # (num_graph,)
        forces: torch.Tensor | None = None,  # (num_node, 3)
        stress: torch.Tensor | None = None,  # (num_graph, 3, 3)
        batch: torch.Tensor | None = None,  # (num_node,)
        sid: list[str] | None = None,
        dataset: list[str] | str | None = None,
    ):
        self.__keys__ = set(_REQUIRED_KEYS)

        # this conversion must have been done somewhere in
        # pytorch geoemtric data
        self.pos = pos
        self.atomic_numbers = atomic_numbers
        self.cell = cell.to(self.pos.dtype)
        self.pbc = pbc
        self.natoms = natoms
        self.edge_index = edge_index
        self.cell_offsets = cell_offsets
        self.nedges = nedges
        self.charge = charge
        self.spin = spin
        self.fixed = fixed
        self.tags = tags
        self.sid = sid if sid is not None else [""]

        # Always normalize dataset to a list for consistent batching
        if dataset is not None:
            if isinstance(dataset, str):
                self.dataset = [dataset]
            elif isinstance(dataset, list):
                self.dataset = dataset
            else:
                raise ValueError(
                    f"dataset must be a string or list of strings, got {type(dataset)}"
                )

        # tagets
        if energy is not None:
            self.energy = energy
        if forces is not None:
            self.forces = forces
        if stress is not None:
            self.stress = stress

        # batch related
        if batch is not None:
            self.batch = batch
        else:
            self.batch = torch.zeros_like(self.atomic_numbers)

        # id
        if isinstance(sid, str):
            self.sid = [sid]
        elif isinstance(sid, list):
            self.sid = sid
        else:
            self.sid = [""]

        self.__slices__ = None
        self.__cumsum__ = None
        self.__cat_dims__ = None
        self.__natoms_list__ = None

        # self.custom_fields = {}

        self.validate()

    @property
    def task_name(self):
        return getattr(self, "dataset", None)

    @task_name.setter
    def task_name(self, value):
        self.dataset = value

    def assign_batch_stats(self, slices, cumsum, cat_dims, natoms_list):
        self.__slices__ = slices
        self.__cumsum__ = cumsum
        self.__cat_dims__ = cat_dims
        self.__natoms_list__ = natoms_list

    def get_batch_stats(self):
        return self.__slices__, self.__cumsum__, self.__cat_dims__, self.__natoms_list__

    def validate(self):
        # shape checks
        assert (
            self.pos.shape[0]
            == self.atomic_numbers.shape[0]
            == self.num_nodes
            == self.natoms.sum().item()
        )
        assert self.pos.dim() == 2
        assert self.pos.shape[1] == 3
        assert self.atomic_numbers.dim() == 1
        assert self.cell.shape[0] == self.pbc.shape[0] == self.num_graphs
        assert self.cell.dim() == 3
        assert self.cell.shape[1:] == (3, 3)
        assert self.pbc.dim() == 2
        assert self.pbc.shape[1] == 3
        assert self.edge_index.shape[0] == 2
        assert self.cell_offsets.shape[0] == self.edge_index.shape[1]
        assert self.nedges.sum().item() == self.edge_index.shape[1]
        assert self.fixed.shape[0] == self.pos.shape[0]
        assert self.tags.shape[0] == self.pos.shape[0]
        assert self.batch.shape == self.atomic_numbers.shape
        assert int(self.batch.max()) + 1 == self.num_graphs
        assert len(self.sid) == self.num_graphs

        if "dataset" in self.__keys__:
            assert isinstance(self.dataset, list), "dataset must always be a list"
            assert len(self.dataset) == self.num_graphs

        # dtype checks
        assert (
            self.pos.dtype == self.cell.dtype == self.cell_offsets.dtype
        ), "Positions, cell, cell_offsets are all expected to be same. Check data going into AtomicData is correct dtype"
        assert self.pos.dtype in (
            torch.float32,
            torch.float64,
        ), "Positions, cell, cell_offsets are all expected to be f32/f64"
        assert self.atomic_numbers.dtype == torch.long
        assert self.edge_index.dtype == torch.long
        assert self.pbc.dtype == torch.bool
        assert self.fixed.dtype == self.tags.dtype == torch.long
        assert self.batch.dtype == torch.long
        assert isinstance(self.num_nodes, int)
        assert isinstance(self.num_graphs, int)

        # NOTE maybe do re matching instead of exact key matching for energy/forces/stress
        if hasattr(self, "energy"):
            assert self.energy.dim() == 1
            assert self.energy.shape[0] == self.num_graphs
            assert self.energy.dtype == torch.float
        if hasattr(self, "forces"):
            assert self.forces.shape[0] == self.pos.shape[0]
            assert self.forces.shape[1] == 3
            assert self.forces.dtype == torch.float
        if hasattr(self, "stress"):
            # NOTE: usually decomposed. for EFS prediction right now we reshape to (9,). need to discuss, perhaps use (1,3,3)
            assert (
                self.stress.dim() == 3
                and self.stress.shape[1:] == (3, 3)
                or (self.stress.dim() == 2 and self.stress.shape[1:] == (9,))
            )
            assert self.stress.shape[0] == self.num_graphs
            assert self.stress.dtype == torch.float

        if self.sid is not None:
            assert isinstance(self.sid, list)
            assert all(isinstance(s, str) for s in self.sid)
            assert len(self.sid) == self.num_graphs

        # for k, v in self.custom_fields.items():
        #     assert self[k].shape == v['shape']
        #     assert self[k].dtype == v['dtype']

    @classmethod
    def from_ase(
        cls,
        input_atoms: ase.Atoms,
        r_edges: bool = False,
        radius: float = 6.0,
        max_neigh: int | None = None,
        sid: str | None = None,
        molecule_cell_size: float | None = None,
        r_energy: bool = True,  # deprecated
        r_forces: bool = True,
        r_stress: bool = True,
        r_data_keys: list[str] | None = None,  # NOT USED, compat for now
        task_name: str | None = None,
        target_dtype: torch.dtype = torch.float32,
    ) -> AtomicData:
        atoms = input_atoms.copy()
        calc = input_atoms.calc
        # TODO: maybe compute a safe cell size if not provided.
        if molecule_cell_size is not None:
            assert (
                atoms.cell.volume == 0.0
            ), "atoms must not have a unit cell to begin with to create a molecule cell"
            # create a molecule box with the molecule centered on it if specified
            atoms.center(vacuum=(molecule_cell_size))
            atoms.pbc = np.array([True, True, True])
        elif np.all(~atoms.pbc):
            pass  # This is fine
        elif not np.all(atoms.pbc) and atoms.cell.volume < 0.1:
            raise ValueError(
                "atoms must either have a cell or have a cell created by setting <molecule_cell_size>."
            )

        atomic_numbers = np.array(atoms.get_atomic_numbers(), copy=True)
        pos = np.array(atoms.get_positions(), copy=True)
        pbc = np.array(atoms.pbc, copy=True)
        cell = np.array(atoms.get_cell(complete=True), copy=True)
        pos = wrap_positions(pos, cell, pbc=pbc, eps=0)

        # wrap positions for CPU graph Generation
        atoms.set_positions(pos)

        atomic_numbers = torch.from_numpy(atomic_numbers).long()
        pos = torch.from_numpy(pos).to(target_dtype)
        pbc = torch.from_numpy(pbc).bool().view(1, 3)
        cell = torch.from_numpy(cell).to(target_dtype).view(1, 3, 3)
        natoms = torch.tensor([pos.shape[0]], dtype=torch.long)

        # graph construction
        if r_edges:
            assert (
                radius is not None
            ), "cutoff must be specified for cpu graph construction."
            assert (
                max_neigh is not None
            ), "max_neigh must be specified for cpu graph construction."
            split_idx_dist = get_neighbors_pymatgen(atoms, radius, max_neigh)
            edge_index, cell_offsets = reshape_features(*split_idx_dist)
            nedges = torch.tensor([edge_index.shape[1]], dtype=torch.long)
        else:
            # empty graph
            edge_index = torch.empty((2, 0), dtype=torch.long)
            cell_offsets = torch.empty((0, 3), dtype=target_dtype)
            nedges = torch.tensor([0], dtype=torch.long)

        # initialized to torch.zeros(natoms) if tags missing.
        # https://wiki.fysik.dtu.dk/ase/_modules/ase/atoms.html#Atoms.get_tags
        tags = torch.from_numpy(atoms.get_tags()).long()
        fixed = torch.zeros(natoms, dtype=torch.long)
        if hasattr(atoms, "constraints"):
            for constraint in atoms.constraints:
                if isinstance(constraint, FixAtoms):
                    fixed[constraint.index] = 1

        if isinstance(calc, (SinglePointCalculator, SinglePointDFTCalculator)):
            results = calc.results
            energy = (
                torch.FloatTensor([results["energy"]]).view(1)
                if "energy" in results
                else None
            )
            forces = (
                torch.FloatTensor(results["forces"]).view(-1, 3)
                if "forces" in results
                else None
            )
            stress = results.get("stress", None)
            if stress is not None and r_stress:
                if stress.shape == (6,):
                    stress = torch.FloatTensor(voigt_6_to_full_3x3_stress(stress)).view(
                        1, 3, 3
                    )
                elif stress.shape in ((3, 3), (9,)):
                    stress = torch.FloatTensor(stress).view(1, 3, 3)
                else:
                    raise ValueError(f"Unknown stress shape, {stress.shape}")
            else:
                stress = None
        else:
            energy = None
            forces = None
            stress = None

        energy = (
            torch.FloatTensor([atoms.info["energy"]])
            if "energy" in atoms.info
            else energy
        )
        forces = (
            torch.FloatTensor(atoms.info["forces"])
            if "forces" in atoms.info
            else forces
        )
        stress = (
            torch.FloatTensor(atoms.info["stress"]).view(1, 3, 3)
            if "stress" in atoms.info
            else stress
        )

        # TODO another way to specify this is to spcify a key. maybe total_charge
        charge = torch.LongTensor(
            [
                atoms.info.get("charge", 0)
                if r_data_keys is not None and "charge" in r_data_keys
                else 0
            ]
        )
        spin = torch.LongTensor(
            [
                atoms.info.get("spin", 0)
                if r_data_keys is not None and "spin" in r_data_keys
                else 0
            ]
        )

        # NOTE: code assumes these are ints.. not tensors
        # charge = atoms.info.get("charge", 0)
        # spin = atoms.info.get("spin", 0)
        data = cls(
            pos=pos,
            atomic_numbers=atomic_numbers,
            cell=cell,
            pbc=pbc,
            natoms=natoms,
            edge_index=edge_index,
            cell_offsets=cell_offsets,
            nedges=nedges,
            charge=tensor_or_int_to_tensor(charge, torch.long),
            spin=tensor_or_int_to_tensor(spin, torch.long),
            fixed=fixed,
            tags=tags,
            energy=energy,
            forces=forces,
            stress=stress,
            sid=[sid] if isinstance(sid, str) else sid,
            dataset=task_name,
        )

        return data

    def to_ase_single(self) -> ase.Atoms:
        assert self.num_graphs == 1, "Data object must contain a single graph."

        atoms = ase.Atoms(
            numbers=self.atomic_numbers.cpu().numpy(),
            positions=self.pos.cpu().numpy(),
            cell=self.cell.squeeze().cpu().numpy(),
            pbc=self.pbc.squeeze().tolist(),
            constraint=FixAtoms(mask=self.fixed.bool().tolist()),
            tags=self.tags.cpu().numpy(),
        )

        if self.__keys__.intersection(["energy", "forces", "stress"]):
            fields = {}
            if hasattr(self, "energy") and self.energy is not None:
                fields["energy"] = self.energy.cpu().numpy()
            if hasattr(self, "forces") and self.forces is not None:
                fields["forces"] = self.forces.cpu().numpy()
            if hasattr(self, "stress") and self.stress is not None:
                if self.stress.shape == (3, 3):
                    fields["stress"] = full_3x3_to_voigt_6_stress(
                        self.stress.squeeze().cpu().numpy()
                    )
                elif self.stress.shape == (6,):
                    fields["stress"] = self.stress.squeeze().numpy()
            atoms.calc = SinglePointCalculator(atoms=atoms, **fields)

        atoms.info = {
            "charge": self.charge.item(),
            "spin": self.spin.item(),
        }

        if self.sid is not None:
            atoms.info["sid"] = self.sid

        return atoms

    def to_ase(self) -> list[ase.Atoms]:
        return [self.get_example(i).to_ase_single() for i in range(self.num_graphs)]

    @classmethod
    def from_dict(cls, dictionary):
        r"""Creates a data object from a python dictionary."""
        assert set(_REQUIRED_KEYS).issubset(
            dictionary.keys()
        ), f"Missing required keys: {set(_REQUIRED_KEYS) - set(dictionary.keys())}"

        data = cls(
            pos=dictionary["pos"],
            atomic_numbers=dictionary["atomic_numbers"],
            cell=dictionary["cell"],
            pbc=dictionary["pbc"],
            natoms=dictionary["natoms"],
            edge_index=dictionary["edge_index"],
            cell_offsets=dictionary["cell_offsets"],
            nedges=dictionary["nedges"],
            charge=dictionary["charge"],
            spin=dictionary["spin"],
            fixed=dictionary.get("fixed", None),
            tags=dictionary.get("tags", None),
            energy=dictionary.get("energy", None),
            forces=dictionary.get("forces", None),
            stress=dictionary.get("stress", None),
            batch=dictionary.get("batch", None),
            sid=dictionary.get("sid", None),
            dataset=dictionary.get("dataset", None),
        )

        # TODO: may require validation for them in the future
        # assign extra keys in dict...
        for key in set(dictionary.keys()) - set(_REQUIRED_KEYS + _OPTIONAL_KEYS):
            data[key] = dictionary[key]

        return data

    # TODO clean
    def to_dict(self):
        return dict(self)

    # TODO clean
    def values(self):
        return [item for _, item in self]

    ###############################
    #       basic operations      #
    ###############################

    @property
    def num_nodes(self) -> int:
        r"""Returns or sets the number of nodes in the graph."""
        return self.pos.size(0)

    @property
    def num_edges(self) -> int:
        """Returns the number of edges in the graph."""
        return self["edge_index"].size(1)

    @property
    def num_graphs(self) -> int:
        """Returns the number of graphs in the batch."""
        return int(self.batch.max()) + 1

    def __len__(self):
        return self.num_graphs

    def get(self, key, default):
        if key in self:
            return self[key]
        return default

    def __getitem__(self, idx):
        if isinstance(idx, str):
            return getattr(self, idx)
        elif isinstance(idx, (int, np.integer)):
            return self.get_example(idx)
        else:
            return self.index_select(idx)

    # TODO: maybe should not allow this. this is currently used a lot though.
    def __setitem__(self, key: str, value: torch.Tensor):
        # TODO call add custom field? if validation checks are given?
        """Sets the attribute :obj:`key` to :obj:`value`."""
        # renaming keys is not supported if this function validate keys.
        # assert key in self.__keys__ or key in set(_REQUIRED_KEYS + _OPTIONAL_KEYS + ['batch']), f"Invalid key: {key}"
        setattr(self, key, value)
        if key not in self.__keys__:
            self.__keys__.add(key)
        # self.validate()

    # TODO: maybe should not allow this. this is currently used a lot though.
    def __setattr__(self, key: str, value: torch.Tensor):
        super().__setattr__(key, value)
        if not key.startswith("__") and key not in self.__keys__:
            self.__keys__.add(key)

    # TODO: maybe should not allow this. this is currently used a lot though.
    def __delitem__(self, key: str):
        """Deletes the attribute :obj:`key`."""
        assert key in self.__keys__, f"Key: <{key}> not found."
        delattr(self, key)
        self.__keys__.remove(key)

    def keys(self):
        return self.__keys__

    def __contains__(self, key):
        r"""Returns :obj:`True`, if the attribute :obj:`key` is present in the
        data."""
        return key in self.__keys__

    def __iter__(self):
        r"""Iterates over all present attributes in the data, yielding their
        attribute names and content."""
        for key in sorted(self.__keys__):
            yield key, self[key]

    def __call__(self, *keys):
        r"""Iterates over all attributes :obj:`*keys` in the data, yielding
        their attribute names and content.
        If :obj:`*keys` is not given this method will iterative over all
        present attributes."""
        for key in keys if keys else sorted(self.__keys__):
            if key in self:
                yield key, self[key]

    def __cat_dim__(self, key, value) -> int:
        r"""Returns the dimension for which :obj:`value` of attribute
        :obj:`key` will get concatenated when creating batches.

        .. note::

            This method is for internal use only, and should only be overridden
            if the batch concatenation process is corrupted for a specific data
            attribute.
        """
        if bool(re.search("index", key)):
            return -1
        return 0

    def __inc__(self, key, value) -> int:
        r"""Returns the incremental count to cumulatively increase the value
        of the next attribute of :obj:`key` when creating batches.

        .. note::

            This method is for internal use only, and should only be overridden
            if the batch concatenation process is corrupted for a specific data
            attribute.
        """
        # Only the `*index*` attribute should be cumulatively summed
        # up when creating batches.
        assert self.num_graphs == 1
        return self.natoms.item() if bool(re.search("index", key)) else 0

    def __apply__(self, item, func):
        if torch.is_tensor(item):
            return func(item)
        elif isinstance(item, (tuple, list)):
            return [self.__apply__(v, func) for v in item]
        elif isinstance(item, dict):
            return {k: self.__apply__(v, func) for k, v in item.items()}
        else:
            return item

    def apply(self, func):
        r"""Applies the function :obj:`func` to all tensor attributes"""
        for key in self.__keys__:
            self[key] = self.__apply__(self[key], func)

        self["batch"] = self.__apply__(self["batch"], func)

        return self

    def contiguous(self):
        r"""Ensures a contiguous memory layout for all tensor attributes"""
        return self.apply(lambda x: x.contiguous())

    def to(self, device, **kwargs):
        r"""Performs tensor dtype and/or device conversion for all tensor attributes"""
        return self.apply(lambda x: x.to(device, **kwargs))

    def cpu(self):
        r"""Copies all tensor attributes to CPU memory."""
        return self.apply(lambda x: x.cpu())

    def cuda(self, device=None, non_blocking=False):
        r"""Copies all tensor attributes to GPU memory."""
        return self.apply(lambda x: x.cuda(device=device, non_blocking=non_blocking))

    def clone(self):
        r"""Performs a deep-copy of the data object."""
        data_dict = {}
        for key in self.__keys__:
            if torch.is_tensor(self[key]):
                data_dict[key] = self[key].clone()
            else:
                # TODO: with this we should stop making sid special.
                data_dict[key] = copy.deepcopy(self[key])
                # ""
                # print(key)
                # raise ValueError("keys must correspond to torch tensors.")
        data_dict["sid"] = copy.deepcopy(self.sid)
        data_dict["batch"] = self.batch.clone()
        batch_stats = copy.deepcopy(self.get_batch_stats())
        cloned = AtomicData.from_dict(data_dict)
        cloned.assign_batch_stats(*batch_stats)

        return cloned

    def __repr__(self):
        cls = str(self.__class__.__name__)
        has_dict = any(isinstance(item, dict) for _, item in self)

        if not has_dict:
            info = [size_repr(key, item) for key, item in self]
            return "{}({})".format(cls, ", ".join(info))
        else:
            info = [size_repr(key, item, indent=2) for key, item in self]
            return "{}(\n{}\n)".format(cls, ",\n".join(info))

    ###############################
    # operations related to batch #
    ###############################

    def get_example(self, idx: int) -> AtomicData:
        r"""
        Reconstructs the :class:`AtomicData` object at index
        :obj:`idx` from a batched AtomicData object.
        """
        if self.num_graphs == 1 and (idx == 0 or idx == -1):
            return self

        data_dict = {}
        idx = self.num_graphs + idx if idx < 0 else idx
        for key in self.__slices__:
            item = self[key]
            if self.__cat_dims__[key] is None:
                # The item was concatenated along a new batch dimension,
                # so just index in that dimension:
                item = item[idx]
            else:
                # Narrow the item based on the values in `__slices__`.
                if isinstance(item, torch.Tensor):
                    dim = self.__cat_dims__[key]
                    start = self.__slices__[key][idx]
                    end = self.__slices__[key][idx + 1]
                    item = item.narrow(dim, start, end - start)
                else:
                    start = self.__slices__[key][idx]
                    end = self.__slices__[key][idx + 1]
                    item = item[start:end]
                    item = item[0] if len(item) == 1 else item

            # Decrease its value by `cumsum` value:
            cum = self.__cumsum__[key][idx]
            if isinstance(item, torch.Tensor):
                if not isinstance(cum, int) or cum != 0:
                    item = item - cum
            elif isinstance(item, (int, float)):
                item = item - cum

            data_dict[key] = item

        data_dict["batch"] = torch.zeros_like(data_dict["atomic_numbers"])
        data_dict["sid"] = [self.sid[idx]]

        if hasattr(self, "dataset"):
            data_dict["dataset"] = self.dataset[idx]

        return AtomicData.from_dict(data_dict)

    def index_select(self, idx: IndexType) -> list[AtomicData]:
        if isinstance(idx, slice):
            idx = list(range(self.num_graphs)[idx])

        elif isinstance(idx, torch.Tensor) and idx.dtype == torch.long:
            idx = idx.flatten().tolist()

        elif isinstance(idx, torch.Tensor) and idx.dtype == torch.bool:
            idx = idx.flatten().nonzero(as_tuple=False).flatten().tolist()

        elif isinstance(idx, Sequence) and not isinstance(idx, str):
            pass

        else:
            raise IndexError(
                f"Only integers, slices (':'), list, tuples, torch.tensor"
                f"are valid indices (got "
                f"'{type(idx).__name__}')"
            )

        return [self.get_example(i) for i in idx]

    def batch_to_atomicdata_list(self) -> list[AtomicData]:
        r"""Reconstructs the list of :class:`torch_geometric.data.Data` objects
        from the batch object.
        The batch object must have been created via :meth:`from_data_list` in
        order to be able to reconstruct the initial objects."""
        return [self.get_example(i) for i in range(self.num_graphs)]

    def update_batch_edges(
        self, edge_index: torch.Tensor, cell_offsets: torch.Tensor, nedges: torch.Tensor
    ) -> AtomicData:
        r"""Update the connectivity of each batched AtomicData sample.

        Args:
            edge_index (torch.Tensor): New batch edge_index (shape [2, total_edges]).
            cell_offsets (torch.Tensor): Cell offsets per edge (shape [total_edges, 3]).
            nedges (torch.Tensor): Number of edges per system (shape [num_systems]).

        Returns:
            AtomicData: The updated batch object.
        """
        self.edge_index = edge_index
        self.cell_offsets = cell_offsets
        self.nedges = nedges
        edge_slices = [0] + torch.cumsum(nedges, dim=0).tolist()
        self.__slices__["edge_index"] = edge_slices
        self.__slices__["cell_offsets"] = edge_slices
        return self


def atomicdata_list_to_batch(
    data_list: list[AtomicData], exclude_keys: Optional[list] = None
) -> AtomicData:
    """
    all data points must be single graphs and have the same set of keys.
    TODO: exclude keys?
    """
    # this does not include the sid/fids
    # arbitrary set of keys handled at set_item?
    if exclude_keys is None:
        exclude_keys = []
    keys = list(set(data_list[0].keys()))

    batched_data_dict = {k: [] for k in keys}
    batch = []

    device = None
    slices = {key: [0] for key in keys}
    cumsum = {key: [0] for key in keys}
    cat_dims = {}
    natoms_list, sid_list = [], []

    for i, data in enumerate(data_list):
        assert (
            data.num_graphs == 1
        ), "data list must only contain single-graph AtomicData objects."

        for key in keys:
            item = data[key]

            # Increase values by `cumsum` value.
            cum = cumsum[key][-1]
            if isinstance(item, torch.Tensor) and item.dtype != torch.bool:
                if not isinstance(cum, int) or cum != 0:
                    item = item + cum
            elif isinstance(item, (int, float)):
                item = item + cum

            # Gather the size of the `cat` dimension.
            size = 1
            cat_dim = data.__cat_dim__(key, data[key])
            # 0-dimensional torch.Tensors have no dimension along which to
            # concatenate, so we set `cat_dim` to `None`.
            if isinstance(item, torch.Tensor) and item.dim() == 0:
                cat_dim = None
            cat_dims[key] = cat_dim

            # Add a batch dimension to items whose `cat_dim` is `None`:
            if isinstance(item, torch.Tensor) and cat_dim is None:
                cat_dim = 0  # Concatenate along this new batch dimension.
                item = item.unsqueeze(0)
                device = item.device
            elif isinstance(item, torch.Tensor):
                size = item.size(cat_dim)
                device = item.device

            batched_data_dict[key].append(item)  # Append item to the attribute list.

            slices[key].append(size + slices[key][-1])
            inc = data.__inc__(key, item)
            if isinstance(inc, (tuple, list)):
                inc = torch.tensor(inc)
            cumsum[key].append(inc + cumsum[key][-1])

        natoms_list.append(data.natoms.item())
        sid_list.extend(data.sid)
        item = torch.full((data.natoms,), i, dtype=torch.long, device=device)
        batch.append(item)

    ref_data = data_list[0]
    for key in keys:
        items = batched_data_dict[key]
        item = items[0]
        cat_dim = ref_data.__cat_dim__(key, item)
        cat_dim = 0 if cat_dim is None else cat_dim
        if torch.is_tensor(item):
            batched_data_dict[key] = torch.cat(items, cat_dim)

        # TODO: this allows non-tensor fields to be batched.
        # we might want to remove support for that.
        else:
            # For list attributes, flatten nested lists to maintain consistency
            # This handles cases where already-batched data is re-batched
            if items and all(isinstance(item, list) for item in items):
                flattened = []
                for item in items:
                    flattened.extend(item)
                batched_data_dict[key] = flattened
            else:
                batched_data_dict[key] = items

    batched_data_dict["batch"] = torch.cat(batch, dim=-1)
    batched_data_dict["sid"] = sid_list
    atomic_data_batch = AtomicData.from_dict(batched_data_dict)
    atomic_data_batch.assign_batch_stats(slices, cumsum, cat_dims, natoms_list)

    return atomic_data_batch.contiguous()


def tensor_or_int_to_tensor(x, dtype=torch.int):
    if isinstance(x, int):
        return torch.Tensor(x, dtype=dtype)
    elif isinstance(x, torch.Tensor):
        assert x.dtype == dtype, "Tensor is not of right dtype"
        return x
    raise ValueError(f"type({x}) is not an int or tensor")



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import torch

EPS = 1e-7


class Safeacos(torch.autograd.Function):
    @staticmethod
    def forward(ctx, x):
        x_clamped = x.clamp(-1 + EPS, 1 - EPS)
        ctx.save_for_backward(x_clamped)
        return torch.acos(x)

    @staticmethod
    def backward(ctx, grad_output):
        (x_clamped,) = ctx.saved_tensors
        denom = torch.sqrt(1 - x_clamped.pow(2)).clamp(min=EPS)
        return -grad_output / denom


class Safeatan2(torch.autograd.Function):
    @staticmethod
    def forward(ctx, y, x):
        ctx.save_for_backward(y, x)
        return torch.atan2(y, x)

    @staticmethod
    @torch.compiler.disable
    def backward(ctx, grad_output):
        y, x = ctx.saved_tensors
        denom = (x.pow(2) + y.pow(2)).clamp(min=EPS)
        return (x / denom) * grad_output, (-y / denom) * grad_output


def init_edge_rot_euler_angles(edge_distance_vec):
    # we need to clamp the output here because if using compile
    # normalize can return >1.0 , pytorch #163082
    xyz = torch.nn.functional.normalize(edge_distance_vec).clamp(-1.0, 1.0)
    x, y, z = torch.split(xyz, 1, dim=1)

    # latitude (beta)
    beta = Safeacos.apply(y.squeeze(-1))

    # longitude (alpha)
    alpha = Safeatan2.apply(x.squeeze(-1), z.squeeze(-1))

    # random gamma (roll)
    gamma = torch.rand_like(alpha) * 2 * torch.pi
    # intrinsic to extrinsic swap
    return -gamma, -beta, -alpha


# Borrowed from e3nn @ 0.4.0:
# https://github.com/e3nn/e3nn/blob/0.4.0/e3nn/o3/_wigner.py#L37
# In 0.5.0, e3nn shifted to torch.matrix_exp which is significantly slower:
# https://github.com/e3nn/e3nn/blob/0.5.0/e3nn/o3/_wigner.py#L92
def wigner_D(
    lv: int,
    alpha: torch.Tensor,
    beta: torch.Tensor,
    gamma: torch.Tensor,
    _Jd: list[torch.Tensor],
) -> torch.Tensor:
    alpha, beta, gamma = torch.broadcast_tensors(alpha, beta, gamma)
    J = _Jd[lv]
    Xa = _z_rot_mat(alpha, lv)
    Xb = _z_rot_mat(beta, lv)
    Xc = _z_rot_mat(gamma, lv)
    return Xa @ J @ Xb @ J @ Xc


def _z_rot_mat(angle: torch.Tensor, lv: int) -> torch.Tensor:
    M = angle.new_zeros((*angle.shape, 2 * lv + 1, 2 * lv + 1))

    # The following code needs to replaced for a for loop because
    # torch.export barfs on outer product like operations
    # ie: torch.outer(frequences, angle) (same as frequencies * angle[..., None])
    # will place a non-sense Guard on the dimensions of angle when attempting to export setting
    # angle (edge dimensions) as dynamic. This may be fixed in torch2.4.

    # inds = torch.arange(0, 2 * lv + 1, 1, device=device)
    # reversed_inds = torch.arange(2 * lv, -1, -1, device=device)
    # frequencies = torch.arange(lv, -lv - 1, -1, dtype=dtype, device=device)
    # M[..., inds, reversed_inds] = torch.sin(frequencies * angle[..., None])
    # M[..., inds, inds] = torch.cos(frequencies * angle[..., None])

    inds = list(range(0, 2 * lv + 1, 1))
    reversed_inds = list(range(2 * lv, -1, -1))
    frequencies = list(range(lv, -lv - 1, -1))
    for i in range(len(frequencies)):
        M[..., inds[i], reversed_inds[i]] = torch.sin(frequencies[i] * angle)
        M[..., inds[i], inds[i]] = torch.cos(frequencies[i] * angle)
    return M


def eulers_to_wigner(
    eulers: torch.Tensor,
    start_lmax: int,
    end_lmax: int,
    Jd: list[torch.Tensor],
) -> torch.Tensor:
    """
    set <rot_clip=True> to handle gradient instability when using gradient-based force/stress prediction.
    """
    alpha, beta, gamma = eulers

    size = int((end_lmax + 1) ** 2) - int((start_lmax) ** 2)
    wigner = torch.zeros(len(alpha), size, size, device=alpha.device, dtype=alpha.dtype)
    start = 0
    for lmax in range(start_lmax, end_lmax + 1):
        block = wigner_D(lmax, alpha, beta, gamma, Jd)
        end = start + block.size()[1]
        wigner[:, start:end, start:end] = block
        start = end

    return wigner



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from __future__ import annotations

from .ase_datasets import AseDBDataset, AseReadDataset, AseReadMultiStructureDataset
from .base_dataset import create_dataset
from .collaters.simple_collater import (
    data_list_collater,
)

__all__ = [
    "AseDBDataset",
    "AseReadDataset",
    "AseReadMultiStructureDataset",
    "create_dataset",
    "data_list_collater",
]



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING, Any, Callable

import torch

if TYPE_CHECKING:
    from pathlib import Path

    from torch.nn import Module
    from torch.utils.data import Dataset


def save_checkpoint(
    state,
    checkpoint_dir: str = "checkpoints/",
    checkpoint_file: str = "checkpoint.pt",
) -> str:
    filename = os.path.join(checkpoint_dir, checkpoint_file)
    torch.save(state, filename)
    return filename


def _load_check_duplicates(config: dict, name: str) -> dict[str, torch.nn.Module]:
    """Attempt to load a single file with normalizers/element references and check config for duplicate targets.

    Args:
        config: configuration dictionary
        name: Name of module to use for logging

    Returns:
        dictionary of normalizer or element reference modules
    """
    modules = {}
    if "file" in config:
        modules = torch.load(config["file"])
        logging.info(f"Loaded {name} for the following targets: {list(modules.keys())}")
    # make sure that element-refs are not specified both as fit and file
    fit_targets = config["fit"]["targets"] if "fit" in config else []
    duplicates = list(
        filter(
            lambda x: x in fit_targets,
            list(config) + list(modules.keys()),
        )
    )
    if len(duplicates) > 0:
        logging.warning(
            f"{name} values for the following targets {duplicates} have been specified to be fit and also read"
            f" from a file. The files read from file will be used instead of fitting."
        )
    duplicates = list(filter(lambda x: x in modules, config))
    if len(duplicates) > 0:
        logging.warning(
            f"Duplicate {name} values for the following targets {duplicates} where specified in the file "
            f"{config['file']} and an explicitly set file. The normalization values read from "
            f"{config['file']} will be used."
        )
    return modules


def _load_from_config(
    config: dict,
    name: str,
    fit_fun: Callable[[list[str], Dataset, Any, ...], dict[str, Module]],
    create_fun: Callable[[str | Path], Module],
    dataset: Dataset,
    checkpoint_dir: str | Path | None = None,
    **fit_kwargs,
) -> dict[str, torch.nn.Module]:
    """Load or fit normalizers or element references from config

    If a fit is done, a fitted key with value true is added to the config to avoid re-fitting
    once a checkpoint has been saved.

    Args:
        config: configuration dictionary
        name: Name of module to use for logging
        fit_fun: Function to fit modules
        create_fun: Function to create a module from file
        checkpoint_dir: directory to save modules. If not given, modules won't be saved.

    Returns:
        dictionary of normalizer or element reference modules

    """
    modules = _load_check_duplicates(config, name)
    for target in config:
        if target == "fit":
            if not config["fit"].get("fitted", False):
                # remove values for output targets that have already been read from files
                targets = [
                    target
                    for target in config["fit"]["targets"]
                    if target not in modules
                ]
                fit_kwargs.update(
                    {k: v for k, v in config["fit"].items() if k != "targets"}
                )
                modules.update(fit_fun(targets=targets, dataset=dataset, **fit_kwargs))
                config["fit"]["fitted"] = True
        # if a single file for all outputs is not provided,
        # then check if a single file is provided for a specific output
        elif target != "file":
            modules[target] = create_fun(**config[target])
        # save the linear references for possible subsequent use
        if checkpoint_dir is not None:
            path = save_checkpoint(
                modules,
                checkpoint_dir,
                f"{name}.pt",
            )
            logging.info(
                f"{name} checkpoint for targets {list(modules.keys())} have been saved to: {path}"
            )

    return modules



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import os
import re
import shutil
from abc import ABCMeta, abstractmethod
from glob import glob
from typing import TYPE_CHECKING, ClassVar, TypeVar

from fairchem.core.components.runner import Runner

if TYPE_CHECKING:
    from collections.abc import Sequence

    from ase.calculators import Calculator


R = TypeVar("R")


class CalculateRunner(Runner, metaclass=ABCMeta):
    """Runner to run calculations/predictions using an ASE-like calculator and save results to file.

    Note:
        When running with the `fairchemv2` cli, the `job_config` and attribute is set at
        runtime to those given in the config file.

    See the Runner interface class for implementation details.

    Attributes:
        job_config (DictConfig): a managed attribute that gives access to the job config
        result_glob_pattern (str): glob pattern of results written to file
    """

    # pattern used in writing result files
    result_glob_pattern: ClassVar[str] = "*"

    def __init__(
        self,
        calculator: Calculator,
        input_data: Sequence,
    ):
        """Initialize the CalculateRunner with a calculator and input data.

        Args:
            calculator (Calculator): ASE-like calculator to perform calculations
            input_data (Sequence): Input data to be processed by the calculator
        """
        self._calculator = calculator
        self._input_data = input_data
        self._already_calculated = False

    @property
    def calculator(self) -> Calculator:
        """Get the calculator instance.

        Returns:
            Calculator: The ASE-like calculator used for calculations
        """
        return self._calculator

    @property
    def input_data(self) -> Sequence:
        """Get the input data.

        Returns:
            Sequence: The input data to be processed
        """
        return self._input_data

    @abstractmethod
    def calculate(self, job_num: int = 0, num_jobs: int = 1) -> R:
        """Run any calculation using an ASE like Calculator.

        Args:
            job_num (int, optional): Current job number in array job. Defaults to 0.
            num_jobs (int, optional): Total number of jobs in array. Defaults to 1.

        Returns:
            R: Results of the calculation
        """
        raise NotImplementedError

    @abstractmethod
    def write_results(
        self, results: R, results_dir: str, job_num: int = 0, num_jobs: int = 1
    ) -> None:
        """Write results to file in results_dir.

        Args:
            results (R): Results from the calculation
            results_dir (str): Directory to write results to
            job_num (int, optional): Current job number in array job. Defaults to 0.
            num_jobs (int, optional): Total number of jobs in array. Defaults to 1.
        """
        raise NotImplementedError

    @abstractmethod
    def save_state(self, checkpoint_location: str, is_preemption: bool = False) -> bool:
        """Save the current state of the calculation to a checkpoint.

        Args:
            checkpoint_location (str): Location to save the checkpoint
            is_preemption (bool, optional): Whether this save is due to preemption. Defaults to False.

        Returns:
            bool: True if state was successfully saved, False otherwise
        """
        raise NotImplementedError

    def load_state(self, checkpoint_location: str | None) -> None:
        """Load a previously saved state from a checkpoint.

        Args:
            checkpoint_location (str | None): Location of the checkpoint to load, or None if no checkpoint
        """
        if checkpoint_location is not None:
            os.makedirs(
                self.job_config.metadata.results_dir, exist_ok=True
            )  # TODO Should we have all these dir created in cli main?

            existing_results_files = glob(
                os.path.join(checkpoint_location, self.result_glob_pattern)
            )
            if len(existing_results_files) == 0:
                return

            job_array_num = self.job_config.metadata.array_job_num
            job_array_total = self.job_config.scheduler.num_array_jobs
            # try to find and load the results file for the given job
            for file in existing_results_files:
                matches = re.findall(r"\d+", file)
                if (
                    matches
                    and int(matches[-1]) == job_array_num
                    and int(matches[-2]) == job_array_total
                ):
                    # copy it over to the new run directory
                    shutil.copy(file, self.job_config.metadata.results_dir)

                    job_state_path = os.path.join(
                        os.path.dirname(file),
                        f"{self.result_glob_pattern.split('_*')[0]}_{job_array_total}_{job_array_num}.txt",
                    )
                    new_path = os.path.join(
                        self.job_config.metadata.results_dir,
                        os.path.basename(file),
                    )
                    with open(job_state_path, "w") as f:
                        f.write(f"{file} moved to {new_path}")

                    os.remove(file)
                    # make sure we can load the file?
                    self._already_calculated = True
                    return

    def run(self):
        """Run the actual calculation and save results.

        Creates the results directory if it doesn't exist, runs the calculation,
        and writes the results to the specified directory.

        Note:
            Re-implementing this method in derived classes is discouraged.
        """
        if not self._already_calculated:
            os.makedirs(
                self.job_config.metadata.results_dir, exist_ok=True
            )  # TODO Should we have all these dir created in cli main?

            results = self.calculate(
                job_num=self.job_config.metadata.array_job_num,
                num_jobs=self.job_config.scheduler.num_array_jobs,
            )
            self.write_results(
                results,
                self.job_config.metadata.results_dir,
                job_num=self.job_config.metadata.array_job_num,
                num_jobs=self.job_config.scheduler.num_array_jobs,
            )



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import pandas as pd
from monty.serialization import loadfn


def load_json_to_df(
    path: str,
    index_name: str | None,
    index_rename: str | None = None,
    sort_index: bool = True,
) -> pd.DataFrame:
    """Read a json file into a pandas DataFrame, optionally reset the index and sort

    Args:
        path: path to json or compressed json file (ie json.gz)
        index_name: name of column to set as index.
        index_rename: if given will rename the index to this
        sort_index: sort the dataframe by its index

    Returns:
        pd.Dataframe
    """

    obj = loadfn(path)
    dataframe = pd.DataFrame(obj)

    if index_name is not None:
        dataframe = dataframe.set_index(index_name)

    if index_rename is not None:
        dataframe.index.name = index_name

    if sort_index:
        dataframe = dataframe.sort_index()

    return dataframe



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import logging
import math
import time
from typing import TYPE_CHECKING, Iterator

import numba as nb
import numpy as np
import torch
from torch.utils.data import Sampler

from fairchem.core.common import gp_utils

if TYPE_CHECKING:
    from fairchem.core.datasets.base_dataset import BaseDataset


@nb.njit
def get_batches(
    natoms_list: np.array,  # List of number of atoms in each sample
    indices: np.array,  # Indices of the samples
    max_atoms: int,  # Maximum number of atoms allowed in a batch
    min_atoms: int,  # Minimum number of atoms allowed in a batch
) -> tuple[list[list[int]], list[int], int]:
    """
    Greedily creates batches from a list of samples with varying numbers of atoms.

    Args:
    natoms_list: Array of number of atoms in each sample.
    indices: Array of indices of the samples.
    max_atoms: Maximum number of atoms allowed in a batch.

    Returns:
    tuple[list[list[int]], list[int], int]:
        A tuple containing a list of batches, a list of the total number of atoms in each batch,
        and the number of samples that were filtered out because they exceeded the maximum number of atoms.
    """

    # Ensure the inputs are valid
    assert max_atoms > 0
    assert len(natoms_list) > 0
    assert len(natoms_list) == len(indices)

    # Initialize variables to keep track of the batches and the total number of atoms in each batch
    batches = []  # List of batches, where each batch is a list of indices
    run_sum = 0  # Running total of the number of atoms in the current batch
    cur_batch = nb.typed.List.empty_list(nb.int64)  # Current batch being constructed
    atom_counts = nb.typed.List.empty_list(
        nb.int64
    )  # List of total number of atoms in each batch
    samples_filtered = 0  # Number of samples filtered out because they exceeded the maximum number of atoms

    # Iterate over the samples
    for idx, atoms in zip(indices, natoms_list):
        # If the sample has too many atoms, filter it out
        if atoms > max_atoms:
            samples_filtered += 1
            continue

        # If adding the sample to the current batch would not exceed the maximum number of atoms, add it
        if run_sum + atoms <= max_atoms:
            cur_batch.append(idx)
            run_sum += atoms
        # Otherwise, start a new batch
        else:
            if run_sum >= min_atoms:
                batches.append(cur_batch)
                atom_counts.append(run_sum)
            cur_batch = nb.typed.List([idx])
            run_sum = atoms

    # Add the last batch
    if run_sum >= min_atoms:
        batches.append(cur_batch)
        atom_counts.append(run_sum)

    # Return the batches, atom counts, and the number of samples filtered out
    return [list(x) for x in batches], list(atom_counts), samples_filtered


class MaxAtomDistributedBatchSampler(Sampler[list[int]]):
    """
    A custom batch sampler that distributes batches across multiple GPUs to ensure efficient training.

    Args:
    dataset (BaseDataset): The dataset to sample from.
    max_atoms (int): The maximum number of atoms allowed in a batch.
    num_replicas (int): The number of GPUs to distribute the batches across.
    rank (int): The rank of the current GPU.
    seed (int): The seed for shuffling the dataset.
    shuffle (bool): Whether to shuffle the dataset. Defaults to True.
    drop_last (bool): Whether to drop the last batch if its size is less than the maximum allowed size. Defaults to False.

    This batch sampler is designed to work with the BaseDataset class and is optimized for distributed training.
    It takes into account the number of atoms in each sample and ensures that the batches are distributed evenly across GPUs.
    """

    def __init__(
        self,
        dataset: BaseDataset,
        max_atoms: int,
        num_replicas: int,
        rank: int,
        seed: int,
        shuffle: bool = True,
        drop_last: bool = False,
        min_atoms: int = 0,
    ) -> None:
        self.dataset = dataset
        self.max_atoms = max_atoms
        self.min_atoms = min_atoms
        self.num_replicas = num_replicas
        self.rank = rank
        assert self.num_replicas > 0
        assert self.rank < self.num_replicas

        if gp_utils.initialized():
            assert (
                min_atoms >= gp_utils.get_gp_world_size()
            ), "Min atoms needs to be at least gp world size!"

        self.seed = seed
        self.shuffle = shuffle
        self.drop_last = drop_last

        self.epoch = 0
        self.start_iter = 0

        # we pre-create the batches here and can only do this once, otherwise everytime we get this iterator we might get a different number of batches
        self.all_batches = self._prepare_batches()
        # If the dataset length is evenly divisible by # of replicas, then there
        # is no need to drop any data, since the dataset will be split equally.
        if self.drop_last and len(self.all_batches) % self.num_replicas != 0:
            # Split to nearest available length that is evenly divisible.
            # This is to ensure each rank receives the same amount of data when
            # using this Sampler.
            self.num_samples = math.ceil(
                (len(self.all_batches) - self.num_replicas) / self.num_replicas
            )
        else:
            self.num_samples = math.ceil(len(self.all_batches) / self.num_replicas)
        self.total_size = self.num_samples * self.num_replicas
        assert (
            len(self.all_batches) >= self.num_replicas
        ), "there are less batches than ranks!"

    def _prepare_batches(self) -> list[int]:
        # shuffle is not optional here since the metadata tend to be sorted by atom count and the resulting batches will be highly uneven
        rng = np.random.default_rng(self.seed)
        original_indices = rng.permutation(len(self.dataset))
        # TODO: this is slow
        t0 = time.time()
        natoms_list = self.dataset.get_metadata("natoms", original_indices.tolist())
        t1 = time.time()
        indices, atoms_count, samples_filtered = get_batches(
            np.array(natoms_list), original_indices, self.max_atoms, self.min_atoms
        )
        t2 = time.time()
        logging.info(
            f"Sampler batch generation times: get natoms: {t1 - t0}, total: {t2 - t0}"
        )
        logging.info(
            f"MaxAtomDistributedSampler generated {len(indices)} batches with total atoms {np.sum(natoms_list)}, max: {max(atoms_count)}, min: {min(atoms_count)}, mean: {np.mean(atoms_count)}, std: {np.std(atoms_count)}"
        )
        logging.info(
            f"{samples_filtered} samples were removed because they exceed {self.max_atoms} atoms"
        )
        return indices

    def __len__(self) -> int:
        return self.num_samples

    def __iter__(self) -> Iterator[list[int]]:
        # based on current rank, return an iterator over batches
        if self.shuffle:
            # deterministically shuffle based on epoch and seed
            g = torch.Generator()
            g.manual_seed(self.seed + self.epoch)
            indices = torch.randperm(len(self.all_batches), generator=g).tolist()  # type: ignore[arg-type]
        else:
            indices = list(range(len(self.all_batches)))  # type: ignore[arg-type]

        if not self.drop_last:
            # add extra samples to make it evenly divisible
            padding_size = self.total_size - len(indices)
            if padding_size <= len(indices):
                indices += indices[:padding_size]
            else:
                indices += (indices * math.ceil(padding_size / len(indices)))[
                    :padding_size
                ]
        else:
            # remove tail of data to make it evenly divisible.
            indices = indices[: self.total_size]

        assert len(indices) == self.total_size

        indices = indices[self.rank : self.total_size : self.num_replicas]
        assert len(indices) == self.num_samples
        # slice of batch indices
        batch_slice = [self.all_batches[i] for i in indices]
        assert (
            self.start_iter < len(batch_slice)
        ), f"starting iteration {self.start_iter} must be less than size of the slice of batches! {len(batch_slice)}"
        return iter(batch_slice[self.start_iter :])

    def set_epoch_and_start_iteration(self, epoch: int, start_iter: int) -> None:
        self.epoch = epoch
        self.start_iter = start_iter



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import json
import os
import time
from abc import ABCMeta, abstractmethod
from typing import Any

from omegaconf import DictConfig

from fairchem.core.components.utils import ManagedAttribute


class Runner(metaclass=ABCMeta):
    """Represents an abstraction over things that run in a loop and can save/load state.

    ie: Trainers, Validators, Relaxation all fall in this category.

    Note:
        When running with the `fairchemv2` cli, the `job_config` and attribute is set at
        runtime to those given in the config file.

    Attributes:
        job_config (DictConfig): a managed attribute that gives access to the job config
    """

    job_config = ManagedAttribute(enforced_type=DictConfig)

    @abstractmethod
    def run(self) -> Any:
        raise NotImplementedError

    @abstractmethod
    def save_state(self, checkpoint_location: str, is_preemption: bool = False) -> bool:
        raise NotImplementedError

    @abstractmethod
    def load_state(self, checkpoint_location: str | None) -> None:
        raise NotImplementedError


class MockRunner(Runner):
    """Used for testing"""

    def __init__(self, x: int, y: int, z: int, sleep_seconds: float = 0.0):
        self.x = x
        self.y = y
        self.z = z
        self.sleep_seconds = sleep_seconds
        self._loaded_state = None

    def run(self) -> Any:
        if self.sleep_seconds > 0:
            elapsed = 0.0
            while elapsed < self.sleep_seconds:
                time.sleep(0.1)
                elapsed += 0.1

        if self.x + self.y > 1000:
            raise ValueError("sum is greater than 1000!")
        return self.x + self.y

    def save_state(self, checkpoint_location: str, is_preemption: bool = False) -> bool:
        os.makedirs(checkpoint_location, exist_ok=True)
        state = {
            "x": self.x,
            "y": self.y,
            "z": self.z,
            "is_preemption": is_preemption,
        }
        with open(os.path.join(checkpoint_location, "mock_state.json"), "w") as f:
            json.dump(state, f)
        return True

    def load_state(self, checkpoint_location: str | None) -> None:
        if checkpoint_location is None:
            return
        state_file = os.path.join(checkpoint_location, "mock_state.json")
        if os.path.exists(state_file):
            with open(state_file) as f:
                self._loaded_state = json.load(f)
                self.x = self._loaded_state["x"]
                self.y = self._loaded_state["y"]
                self.z = self._loaded_state["z"]



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import argparse
import glob
import logging
import os

import fairchem.core

"""
This script provides users with an automated way to download, preprocess (where
applicable), and organize data to readily be used by the existing config files.
"""

DOWNLOAD_LINKS_s2ef: dict[str, dict[str, str]] = {
    "s2ef": {
        "200k": "https://dl.fbaipublicfiles.com/opencatalystproject/data/s2ef_train_200K.tar",
        "2M": "https://dl.fbaipublicfiles.com/opencatalystproject/data/s2ef_train_2M.tar",
        "20M": "https://dl.fbaipublicfiles.com/opencatalystproject/data/s2ef_train_20M.tar",
        "all": "https://dl.fbaipublicfiles.com/opencatalystproject/data/s2ef_train_all.tar",
        "val_id": "https://dl.fbaipublicfiles.com/opencatalystproject/data/s2ef_val_id.tar",
        "val_ood_ads": "https://dl.fbaipublicfiles.com/opencatalystproject/data/s2ef_val_ood_ads.tar",
        "val_ood_cat": "https://dl.fbaipublicfiles.com/opencatalystproject/data/s2ef_val_ood_cat.tar",
        "val_ood_both": "https://dl.fbaipublicfiles.com/opencatalystproject/data/s2ef_val_ood_both.tar",
        "test": "https://dl.fbaipublicfiles.com/opencatalystproject/data/s2ef_test_lmdbs.tar.gz",
        "rattled": "https://dl.fbaipublicfiles.com/opencatalystproject/data/s2ef_rattled.tar",
        "md": "https://dl.fbaipublicfiles.com/opencatalystproject/data/s2ef_md.tar",
    },
}

DOWNLOAD_LINKS_is2re: dict[str, str] = {
    "is2re": "https://dl.fbaipublicfiles.com/opencatalystproject/data/is2res_train_val_test_lmdbs.tar.gz",
}

S2EF_COUNTS = {
    "s2ef": {
        "200k": 200000,
        "2M": 2000000,
        "20M": 20000000,
        "all": 133934018,
        "val_id": 999866,
        "val_ood_ads": 999838,
        "val_ood_cat": 999809,
        "val_ood_both": 999944,
        "rattled": 16677031,
        "md": 38315405,
    },
}


def get_data(datadir: str, task: str, split: str | None, del_intmd_files: bool) -> None:
    os.makedirs(datadir, exist_ok=True)

    if task == "s2ef" and split is None:
        raise NotImplementedError("S2EF requires a split to be defined.")

    download_link: str | None = None
    if task == "s2ef":
        assert split is not None, "Split must be defined for the s2ef dataset task"
        assert (
            split in DOWNLOAD_LINKS_s2ef[task]
        ), f'S2EF "{split}" split not defined, please specify one of the following: {list(DOWNLOAD_LINKS_s2ef["s2ef"].keys())}'
        download_link = DOWNLOAD_LINKS_s2ef[task][split]
    elif task == "is2re":
        download_link = DOWNLOAD_LINKS_is2re[task]
    else:
        raise Exception(f"Unrecognized task {task}")
    assert download_link is not None

    os.system(f"wget {download_link} -P {datadir}")
    filename = os.path.join(datadir, os.path.basename(download_link))
    logging.info("Extracting contents...")
    os.system(f"tar -xvf {filename} -C {datadir}")
    dirname = os.path.join(
        datadir,
        os.path.basename(filename).split(".")[0],
    )
    if task == "s2ef" and split != "test":
        assert split is not None, "Split must be defined for the s2ef dataset task"
        compressed_dir = os.path.join(dirname, os.path.basename(dirname))
        if split in ["200k", "2M", "20M", "all", "rattled", "md"]:
            output_path = os.path.join(datadir, task, split, "train")
        else:
            output_path = os.path.join(datadir, task, "all", split)
        uncompressed_dir = uncompress_data(compressed_dir)
        preprocess_data(uncompressed_dir, output_path)

        verify_count(output_path, task, split)
    if task == "s2ef" and split == "test":
        os.system(f"mv {dirname}/test_data/s2ef/all/test_* {datadir}/s2ef/all")
    elif task == "is2re":
        os.system(f"mv {dirname}/data/is2re {datadir}")

    if del_intmd_files:
        cleanup(filename, dirname)


def uncompress_data(compressed_dir: str) -> str:
    import uncompress

    parser = uncompress.get_parser()
    args, _ = parser.parse_known_args()
    args.ipdir = compressed_dir
    args.opdir = os.path.dirname(compressed_dir) + "_uncompressed"
    uncompress.main(args)
    return args.opdir


def preprocess_data(uncompressed_dir: str, output_path: str) -> None:
    import preprocess_ef as preprocess

    parser = preprocess.get_parser()
    args, _ = parser.parse_known_args()
    args.data_path = uncompressed_dir
    args.out_path = output_path
    preprocess.main(args)


def verify_count(output_path: str, task: str, split: str) -> None:
    paths = glob.glob(os.path.join(output_path, "*.txt"))
    count = 0
    for path in paths:
        with open(path) as fp:
            lines = fp.read().splitlines()
        count += len(lines)
    assert (
        count == S2EF_COUNTS[task][split]
    ), f"S2EF {split} count incorrect, verify preprocessing has completed successfully."


def cleanup(filename: str, dirname: str) -> None:
    import shutil

    if os.path.exists(filename):
        os.remove(filename)
    if os.path.exists(dirname):
        shutil.rmtree(dirname)
    if os.path.exists(dirname + "_uncompressed"):
        shutil.rmtree(dirname + "_uncompressed")


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--task", type=str, help="Task to download")
    parser.add_argument(
        "--split", type=str, help="Corresponding data split to download"
    )
    parser.add_argument(
        "--keep",
        action="store_true",
        help="Keep intermediate directories and files upon data retrieval/processing",
    )
    # Flags for S2EF train/val set preprocessing:
    parser.add_argument(
        "--get-edges",
        action="store_true",
        help="Store edge indices in LMDB, ~10x storage requirement. Default: compute edge indices on-the-fly.",
    )
    parser.add_argument(
        "--num-workers",
        type=int,
        default=1,
        help="No. of feature-extracting processes or no. of dataset chunks",
    )
    parser.add_argument(
        "--ref-energy", action="store_true", help="Subtract reference energies"
    )
    parser.add_argument(
        "--data-path",
        type=str,
        default=os.path.join(os.path.dirname(fairchem.core.__path__[0]), "data"),
        help="Specify path to save dataset. Defaults to 'fairchem.core/data'",
    )

    args: argparse.Namespace
    args, _ = parser.parse_known_args()
    get_data(
        datadir=args.data_path,
        task=args.task,
        split=args.split,
        del_intmd_files=not args.keep,
    )



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

from concurrent.futures import ThreadPoolExecutor
from functools import cached_property
from multiprocessing import cpu_count
from typing import Literal, Protocol

from fairchem.core.units.mlip_unit._batch_serve import setup_batch_predict_server
from fairchem.core.units.mlip_unit.predict import (
    BatchServerPredictUnit,
    MLIPPredictUnit,
)


class ExecutorProtocol(Protocol):
    def submit(self, fn, *args, **kwargs): ...
    def map(self, fn, *iterables, **kwargs): ...
    def shutdown(self, wait: bool = True): ...


def _get_concurrency_backend(
    backend: Literal["threads"], options: dict
) -> ExecutorProtocol:
    """Get a backend to run ASE calculations concurrently."""
    if backend == "threads":
        return ThreadPoolExecutor(**options)
    raise ValueError(f"Invalid concurrency backend: {backend}")


class InferenceBatcher:
    """Batches incoming inference requests."""

    def __init__(
        self,
        predict_unit: MLIPPredictUnit,
        max_batch_size: int = 512,
        batch_wait_timeout_s: float = 0.1,
        num_replicas: int = 1,
        concurrency_backend: Literal["threads"] = "threads",
        concurrency_backend_options: dict | None = None,
        ray_actor_options: dict | None = None,
    ):
        """
        Args:
            predict_unit: The predict unit to use for inference.
            max_batch_size: Maximum number of atoms in a batch.
                The actual number of atoms will likely be larger than this as batches
                are split when num atoms exceeds this value.
            batch_wait_timeout_s: The maximum time to wait for a batch to be ready.
            num_replicas: The number of replicas to use for inference.
            concurrency_backend: The concurrency backend to use for inference.
            concurrency_backend_options: Options to pass to the concurrency backend.
            ray_actor_options: Options to pass to the Ray actor running the batch server.
        """
        self.predict_unit = predict_unit
        self.max_batch_size = max_batch_size
        self.batch_wait_timeout_s = batch_wait_timeout_s
        self.num_replicas = num_replicas

        self.predict_server_handle = setup_batch_predict_server(
            predict_unit=self.predict_unit,
            max_batch_size=self.max_batch_size,
            batch_wait_timeout_s=self.batch_wait_timeout_s,
            num_replicas=self.num_replicas,
            ray_actor_options=ray_actor_options or {},
        )

        if concurrency_backend_options is None:
            concurrency_backend_options = {}

        if (
            concurrency_backend == "threads"
            and "max_workers" not in concurrency_backend_options
        ):
            concurrency_backend_options["max_workers"] = min(cpu_count(), 16)

        self.executor: ExecutorProtocol = _get_concurrency_backend(
            concurrency_backend, concurrency_backend_options
        )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shutdown()

    @cached_property
    def batch_predict_unit(self) -> BatchServerPredictUnit:
        return BatchServerPredictUnit(
            server_handle=self.predict_server_handle,
            predict_unit=self.predict_unit,
        )

    def shutdown(self, wait: bool = True) -> None:
        """Shutdown the executor.

        Args:
            wait: If True, wait for pending tasks to complete before returning.
        """
        if hasattr(self, "executor"):
            self.executor.shutdown(wait=wait)

    def __del__(self):
        """Cleanup on deletion."""
        self.shutdown(wait=False)
